#!/usr/bin/python
# -*- coding: utf-8 -*-

#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os.path
import math
import sys
import re
import gzip
import traceback
import cPickle
import Owlcat.Tables

flagger = parser = ms = msname = None;

def error (message):
  sys.stderr.write("%s: %s\n"%(os.path.basename(sys.argv[0]),message));
  sys.exit(1);

def get_ms (readonly=True):
  global ms;
  global msname;
  if not ms:
    ms = Owlcat.table(msname,readonly=readonly);
  return ms;

def shape_str (label,arr):
  return "%s: %s"%(label,"x".join(map(str,arr.shape)) if arr is not None else "None");

def parse_subset_options (options):
  global parser;
  global flagger;
  global msname;
  global ms;
  subset = {};
  from Owlcat import Parsing

  # DDID and FIELD_ID
  if options.ddid is not None:
    try:
      subset['ddid'] = map(int,options.ddid.split(","));
      print "  ===> DATA_DESC_ID:",subset['ddid'];
    except:
      parser.error("Invalid -D/--ddid option");
  if options.field is not None:
    try:
      subset['fieldid'] = map(int,options.field.split(","));
      print "  ===> FIELD_ID:",subset['fieldid'];
    except:
      parser.error("Invalid -F/--field option");
  # taql
  taqls = [];
  if options.taql:
    taqls.append(options.taql);
    print "  ===> TaQL query:",options.taql;
  # channels
  if options.channels:
    subset['channels'] = map(Parsing.parse_slice,options.channels.split(","));
    print "  ===> channels:",subset['channels'];
  # corr list
  if options.corrs is not None:
    try:
      subset['corrs'] = map(int,options.corrs.split(','));
      print "  ===> correlations:",subset['corrs'];
    except:
      parser.error("Invalid -X/--corrs option");
  # station list
  if options.stations is not None:
    try:
      subset['antennas'] = map(int,options.stations.split(','));
      print "  ===> stations:",subset['antennas'];
    except:
      parser.error("Invalid -S/--stations option");
  # IFR set
  if options.ifrs is not None:
    import Meow.IfrSet
    ifrset = Meow.IfrSet.from_ms(get_ms());
    # print help and exit
    if options.ifrs == "help":
      # print help string, but trim away RTF tags
      print re.sub("<[^>]+>","",ifrset.subset_doc).replace("&lt;","<").replace("&gt;",">");
      sys.exit(0);
    try:
      ifrset = ifrset.subset(options.ifrs);
      print "  ===> ifrs:"," ".join([ifrset.ifr_label(ip,iq) for (ip,p),(iq,q) in ifrset.ifr_index()]);
      if not ifrset.ifrs():
        return None;
    except:
      parser.error("Invalid -I/--ifrs option");
    taqls.append(ifrset.taql_string());
  # clipping
  subset['data_column'] = options.data_column;
  if options.nan:
    subset['data_nan'] = True;
    print "  ===> select %s = NAN or INF"%options.data_column;
  if options.above is not None:
    subset['data_above'] = options.above;
    print "  ===> select |%s|>%f"%(options.data_column,options.above);
  if options.below is not None:
    subset['data_below'] = options.below;
    print "  ===> select |%s|<%f"%(options.data_column,options.below);
  if options.fm_above is not None:
    subset['data_fm_above'] = options.fm_above;
    print "  ===> select mean|%s|>%f"%(options.data_column,options.fm_above);
  if options.fm_below is not None:
    subset['data_fm_below'] = options.fm_below;
    print "  ===> select mean|%s|<%f"%(options.data_column,options.fm_below);
  # join taql queries
  if taqls:
    subset['taql'] = "( " + " ) && ( ".join(taqls) + " )";
  # fill flag args
  for opt in 'data_flagmask','flagmask','flagmask_all','flagmask_none':
    subset[opt] = getattr(options,opt);
  return subset;


if __name__ == "__main__":

  # setup some standard command-line option parsing
  #
  from optparse import OptionParser,OptionGroup
  parser = OptionParser(usage="""%prog: [actions] [options] MS""",
      description="Manipulates flags (bitflags and legacy FLAG/FLAG_ROW columns) in the MS. "
      "Use the selection options to narrow down a subset of the data, and use the action options "
      "to change flags within that subset. Without any action options, statistics on the current "
      "selection are printed -- this is useful as a preview of your intended action."
  );

  group = OptionGroup(parser,"Selection by subset");
  group.add_option("-L","--channels",type="string",
                    help="channel selection: single number or start:end[:step] to select channels start through end-1, "
                    "or start~end[:step] to select channels start through end, with an optional stepping.");
  group.add_option("-T","--timeslots",type="string",
                    help="timeslot selection: single number or start:end to select timeslots start through end-1, "
                    "or start~end to select timeslots start through end.");
  group.add_option("-M","--timeslot-multiplier",type="int",default=1,
                    help="multiplies the timeslot numbers given to -T by the given factor. Default is 1.");
  group.add_option("-X","--corrs",type="string",
                    help="correlation selection. Use comma-separated list of correlation indices.");
  group.add_option("-S","--stations",type="string",
                    help="station (=antenna) selection. Use comma-separated list of station indices."),
  group.add_option("-I","--ifrs",type="string",
                    help="interferometer selection. Use \"-I help\" to get help on selecting ifrs.");
  group.add_option("-D","--ddid",type="string",
                    help="DATA_DESC_ID selection. Single number, or comma-separated list.");
  group.add_option("-F","--field",type="string",
                    help="FIELD_ID selection. Single number, or comma-separated list.");
  group.add_option("-Q","--taql",dest="taql",type="str",
                    help="additional TaQL selection to restrict subset.");
  parser.add_option_group(group);

  group = OptionGroup(parser,"Selection by data value");
  group.add_option("--above",metavar="X",type="float",
                    help="select on abs(data)>X");
  group.add_option("--below",metavar="X",type="float",
                    help="select on abs(data)<X");
  group.add_option("--nan",action="store_true",
                    help="select on invalid data (NaN or infinite)");
  group.add_option("--fm-above",metavar="X",type="float",
                    help="select on mean(abs(data))>X, where mean is over frequencies");
  group.add_option("--fm-below",metavar="X",type="float",
                    help="select on mean(abs(data))<X, where mean is over frequencies");
  group.add_option("-C","--data-column",metavar="COLUMN",type="string",
                    help="data column for --above/--below/--nan options. Default is %default.");
  group.add_option("--data-flagmask",metavar="FLAGS",type="string",
                    help="flags to apply to data column (when e.g. computing mean). Default is %default. See below for "
                    "details on specifying flags.");
  parser.add_option_group(group);

  group = OptionGroup(parser,"Selection by current flags");
  group.add_option("-Y","--flagged-any",metavar="FLAGS",dest='flagmask',type="string",
                    help="selects if any of the specified flags are raised. For this and all other options taking "
                    "a FLAGS argument, FLAGS can be a flagset name or an integer bitmask "
                    "(if bitflags are in use -- see also the -l/--list option). Prefix the bitmask by '0x' to use hex. "
                    "Append a '+L' to include legacy boolean FLAG/FLAG_ROW columns. Use 'all' for "
                    "all bitflags, and 'ALL' for all bitflags plus legacy flags (equivalent to 'all+L'). FLAGS may "
                    "also be a comma-separated list of any of the above terms.");
  group.add_option("-A","--flagged-all",metavar="FLAGS",dest='flagmask_all',type="string",
                    help="selects if all of the specified flags are raised");
  group.add_option("-N","--flagged-none",metavar="FLAGS",dest='flagmask_none',type="string",
                    help="selects if none of the specified flags are raised");
  parser.add_option_group(group);

  group = OptionGroup(parser,"Actions to take on selection (may be combined)");
  group.add_option("-x","--extend-all-corr",action="store_true",
                   help="apply selection to all correlations if at least one is selected");
  group.add_option("-f","--flag",metavar="FLAGS",type="string",
                  help="raise the specified FLAGS");
  group.add_option("-u","--unflag",metavar="FLAGS",type="string",
                  help="clear the specified flags");
  group.add_option("-g","--fill-legacy",metavar="FLAGS",type="string",
                  help="fills legacy FLAG/FLAG_ROW columns using the specified FLAGS. When -f/--flag or -u/--unflag "
                  "or -r/--remove is used, legacy flags are implicitly reset using all bitflags: use '-g -' "
                  "to skip this step. You may also use this option on its own to reset legacy flags (within the "
                  "specified data subset) using some bitmask. Use '-g 0' to clear legacy flags.");
  group.add_option("-c","--create",action="store_true",
                  help="for -f/--flag option only: if a named flagset doesn't exist, creates "
                  "it. Without this option, an error is reported.");
  parser.add_option_group(group);

  group = OptionGroup(parser,"Other options");
  group.add_option("-l","--list",action="store_true",
                  help="lists various info about the MS, including its flagsets.");
  group.add_option("-s","--stats",action="store_true",
                  help="prints per-flagset flagging stats.");
  group.add_option("-r","--remove",metavar="FLAGSET(s)",type="string",
                  help="unflags and removes named flagset(s). You can use a comma-separated list.");
  group.add_option("--export",type="string",metavar="FILENAME",
                  help="exports all flags to flag file. FILENAME may end with .gz to produce a gzip-compressed file. If any flagging actions are specified, these will be done before the export." );
  group.add_option("--import",type="string",dest="_import",metavar="FILENAME",
                  help="imports flags from flag file. If any flagging actions are specified, these will be done after the import.");
  group.add_option("-v","--verbose",metavar="LEVEL",type="int",
                    help="verbosity level for messages. Higher is more verbose, default is 0.");
  group.add_option("--timestamps",action="store_true",
                  help="adds timestamps to verbosity messages.");
  group.add_option("-z","--chunk-size",metavar="NROWS",type="int",default=200000,
                    help="Number of rows to process at once. Default is %default. Set to higher values if you have RAM to spare.");
  parser.add_option_group(group);

  parser.set_defaults(data_column="CORRECTED_DATA",data_flagmask="ALL",
    flagged_any=None,flaged_all=None,flagged_none=None,
    flag=None,unflag=None,fill_legacy=None,
    verbose=0);

  # parse args
  (options,args) = parser.parse_args();
  if len(args) != 1:
    parser.error("Incorrect number of arguments. Use '-h' for help.");
  msname  = args[0];

  import Owlcat

  # import flags from file, if so specified
  # these are the columns that are imported and exported
  FLAGCOLS = "FLAG","FLAG_ROW","BITFLAG","BITFLAG_ROW";
  if options._import:
    try:
      dump = Owlcat.Tables.TableDump(options._import,compress=True);
      ms = get_ms(readonly=False);
      print "Importing flags from %s:"%options._import;
      dump.load(ms,verbose=True);
      dump.close();
      ms.close();
    except:
      traceback.print_exc();
      error("Error importing flags from %s, exiting"%options._import);
    print "Flags imported OK.";

  # if no other actions supplied, enable stats (unless flags were imported, in which case just exit)
  if not (options.flag or options.unflag or options.fill_legacy):
    if options._import:
      sys.exit(0);
    statonly = True;
  else:
    statonly = False;

  import numpy
  import numpy.ma
  import Owlcat.Flagger
  from Owlcat.Flagger import Flagger

  # now, skip most of the actions below if we're in statonly mode and exporting
  if not (statonly and options.export):
    # create flagger object
    flagger = Flagger(msname,verbose=options.verbose,timestamps=options.timestamps,chunksize=options.chunk_size);

    #
    # -l/--list: list MS info
    #
    if options.list:
      ms = get_ms();
      ants = Owlcat.table(ms.getkeyword('ANTENNA')).getcol('NAME');
      ddid_tab = Owlcat.table(ms.getkeyword('DATA_DESCRIPTION'));
      spwids = ddid_tab.getcol('SPECTRAL_WINDOW_ID');
      polids = ddid_tab.getcol('POLARIZATION_ID');
      corrs = Owlcat.table(ms.getkeyword('POLARIZATION')).getcol('CORR_TYPE');
      spw_tab = Owlcat.table(ms.getkeyword('SPECTRAL_WINDOW'));
      ref_freq = spw_tab.getcol('REF_FREQUENCY');
      nchan = spw_tab.getcol('NUM_CHAN');
      fields = Owlcat.table(ms.getkeyword('FIELD')).getcol('NAME');

      print "===> MS is %s"%msname;
      print "  %d antennas: %s"%(len(ants)," ".join(ants));
      print "  %d DATA_DESC_ID(s): "%len(spwids);
      for i,(spw,pol) in enumerate(zip(spwids,polids)):
        print "    %d: %.3f MHz, %d chans x %d correlations"%(i,ref_freq[spw]*1e-6,nchan[spw],len(corrs[pol,:]));
      print "  %d field(s): %s"%(len(fields),", ".join(["%d: %s"%ff for ff in enumerate(fields)]));
      if not flagger.has_bitflags:
        print "No BITFLAG/BITFLAG_ROW columns in this MS. Use the 'addbitflagcol' command to add them.";
      else:
        names = flagger.flagsets.names();
        if names:
          print "  %d flagset(s): "%len(names);
          for name in names:
            mask = flagger.flagsets.flagmask(name);
            print "    '%s': %d (0x%02X)"%(name,mask,mask);
        else:
          print "  No flagsets.";
      print "";
      if options.flag or options.unflag or options.fill_legacy or options.remove:
        print "-l/--list was in effect, so all other options were ignored.";
      sys.exit(0);

    # --flag/--unflag/--remove implies '-g all' by default, '-g -' skips the fill-legacy step
    if options.flag or options.unflag or options.remove:
      if options.fill_legacy is None:
        options.fill_legacy = 'all';
      elif options.fill_legacy == '-':
        options.fill_legacy = None;

    # if no other actions supplied, enable stats (unless flags were imported, in which case just exit)
    if not (options.flag or options.unflag or options.fill_legacy):
      if options._import:
        sys.exit(0);
      statonly = not options.export;
    else:
      statonly = False;

    # convert all the various FLAGS to flagmasks (or Nones)
    for opt in 'data_flagmask','flagmask','flagmask_all','flagmask_none','flag','unflag','fill_legacy':
      value = getattr(options,opt);
      try:
        flagmask = flagger.lookup_flagmask(value,create=(opt=='flag' and options.create));
      except Exception,exc:
        msg = str(exc);
        if opt == 'flag' and not options.create:
          msg += "\nPerhaps you forgot the -c/--create option?";
        error(msg);
      setattr(options,opt,flagmask);

    # clear the legacy flag itself from fill_legacy, otherwise it can have no effect
    if options.fill_legacy is not None:
      options.fill_legacy &= ~Flagger.LEGACY;

    #
    # -r/--remove: remove flagsets
    #
    if options.remove is not None:
      if options.flag or options.unflag:
        error("Can't combine -r/--remove with --f/--flag or -u/--unflag.");
      # get list of names and flagmask to remove
      remove_names = options.remove.split(",");
      names_found = [];
      names_not_found = [];
      flagmask = 0;
      for name in remove_names:
        try:
          flagmask |= flagger.flagsets.flagmask(name);
          names_found.append(name);
        except:
          names_not_found.append(name);
      if not names_found:
        #if names_not_found:
        # error("No such flagset(s): %s"%",".join(names_not_found));
        print "===> WARNING: no flagsets to remove, exiting";
        sys.exit(0);
      # unflag flagsets
      if names_not_found:
        print "===> WARNING: flagset(s) not found, ignoring: %s"%",".join(names_not_found);
      print "===> removing flagset(s) %s"%",".join(names_found);
      print "===> and clearing corresponding flagmask %s"%Flagger.flagmaskstr(flagmask);
      if options.fill_legacy is not None:
        print "===> and filling FLAG/FLAG_ROW using flagmask %s"%Flagger.flagmaskstr(options.fill_legacy);
      flagger.xflag(unflag=flagmask,fill_legacy=options.fill_legacy);
      flagger.flagsets.remove_flagset(*names_found);
      sys.exit(0);

    # parse subset options
    subset = parse_subset_options(options);
    if not subset:
      print "===> ended up with empty subset, exiting";
      sys.exit(0);

    # convert timeslots to reltime option, if specified
    if options.timeslots:
      from Owlcat import Parsing
      tslice = Parsing.parse_slice(options.timeslots,options.timeslot_multiplier);
      times = sorted(set(get_ms().getcol('TIME')));
      time0 = times[0] if tslice.start is None else times[tslice.start];
      time1 = times[-1] if tslice.stop is None else times[tslice.stop-1];
      time0 -= times[0];
      time1 -= times[0];
      subset['reltime'] = time0,time1;
      print "  ===> select timeslots %s (reltime %g~%g s)"%(tslice,time0,time1);

    # at this stage all remaining options are handled the same way
    flagstr = unflagstr = legacystr = None;
    if options.flag is not None:
      flagstr = flagger.flagmaskstr(options.flag);
      print "===> flagging with flagmask %s"%flagstr;
    if options.unflag is not None:
      unflagstr = flagger.flagmaskstr(options.unflag);
      print "===> unflagging with flagmask %s"%unflagstr;
    if options.fill_legacy is not None:
      legacystr = flagger.flagmaskstr(options.fill_legacy);
      print "===> filling legacy flags with flagmask %s"%legacystr;


    # if --stats in effect, loop over all flagsets and print stats
    if options.stats:
      print "===> --stats in effect, showing per-flagset statistics"
      printed_header = False;
      for flagset in list(flagger.flagsets.names())+["+L"]:
        # get stats for this flagset
        subset['flagmask'] = flagger.lookup_flagmask(flagset);
        totrows,sel_nrow,sel_nvis,nvis_A,nvis_B,nvis_C = flagger.xflag(**subset);
        percent = 100.0/sel_nvis if sel_nvis else 0;
        # print them
        if flagset is "+L":
          label = "legacy FLAG/FLAG_ROW";
        else:
          label =  "Flagset %s (0x%02X)"%(flagset,flagger.flagsets.flagmask(flagset));
        if not printed_header:
          printed_header = True;
          rpc = 100.0/totrows if totrows else 0;
          print "===>   MS size:               %8d rows"%totrows;
          print "===>   Data/time selection:   %8d rows, %10d visibilities (%.3g%% of MS rows)"%(sel_nrow,sel_nvis,sel_nrow*rpc);
          if options.channels or options.corrs:
            print "===>   Chan/corr slicing reduces this to    %12d visibilities (%.3g%% of selection)"%(nvis_A,nvis_A*percent);
        print "===>   %-29s includes %10d visibilities (%.3g%% of selection)"%(label,nvis_B,nvis_B*percent);
      sys.exit(0);

    # else not stats mode, do the actual flagging job
    totrows,sel_nrow,sel_nvis,nvis_A,nvis_B,nvis_C = \
      flagger.xflag(flag=options.flag,unflag=options.unflag,fill_legacy=options.fill_legacy,
        flag_allcorr=options.extend_all_corr,
        **subset);
      
    # print stats
    if statonly:
      print "===> No actions were performed. Showing the result of your selection:"
    else:
      print "===> Flagging stats:";
    rpc = 100.0/totrows if totrows else 0;
    print "===>   MS size:               %8d rows"%totrows;
    print "===>   Data/time selection:   %8d rows, %10d visibilities (%.3g%% of MS rows)"%(sel_nrow,sel_nvis,sel_nrow*rpc);
    if legacystr:
      print "===>     (over which legacy flags were filled using flagmask %s)"%legacystr;

    percent = 100.0/sel_nvis if sel_nvis else 0;
    if options.channels or options.corrs:
      print "===>   Chan/corr slicing reduces this to     %10d visibilities (%.3g%% of selection)"%(nvis_A,nvis_A*percent);
    if not (options.flagmask is None and options.flagmask_all is None and options.flagmask_none is None):
      print "===>   Flag selection reduces this to        %10d visibilities (%.3g%% of selection)"%(nvis_B,nvis_B*percent);
    if options.nan or options.above is not None or options.below is not None or \
        options.fm_above is not None or options.fm_below is not None:
      print "===>   Data selection reduces this to         %10d visibilities (%.3g%% of selection)"%(nvis_C,nvis_C*percent);
    if unflagstr:
      print "===>     (which were unflagged using flagmask %s)"%unflagstr;
    if flagstr:
      print "===>     (which were flagged using flagmask %s)"%flagstr;

    flagger.close();

  # export flags from file, if so specified
  if options.export:
    try:
      dump = Owlcat.Tables.TableDump(options.export,write=True,compress=True);
      ms = get_ms();
      colnames = set(ms.colnames());
      print "Exporting flags to %s, %d rows:"%(options.export,ms.nrows());
      for colname in FLAGCOLS:
        if colname in colnames:
          dump.dump_column(ms,colname,verbose=True);
      ms.close();
      dump.close();
    except:
      traceback.print_exc();
      error("Error exporting flags to %s"%options.export);
    print "Flags exported OK.";



# -*- coding: utf-8 -*-


"""This is a library of "funklet operations".
Each function here is compatible with ParmTab.apply().
Input argument is a ParmTab.FunkSlie object.
Return value is a list of output funklets.
""";

import copy

from Timba.Meq import meq
from Timba import dmi
from Timba import mequtils


def average (funkslice):
  """Reduction function to replace all funklets in a slice with their mean.
  This is the canonical example of a reduction function.
  """
  outfunk = funkslice[0];
  shape = getattr(outfunk.coeff,'shape',None);   # shape will be None when coeff is just a float
  # accumulate sum of coefficients and take mean
  for funk in funkslice[1:]:
    outfunk.coeff += funk.coeff;
    # adjust domain to envelope
    for axis in funkslice.slice_axes:
      outfunk.domain[axis] = (min(outfunk.domain[axis][0],funk.domain[axis][0]),
                              max(outfunk.domain[axis][1],funk.domain[axis][1]));
  outfunk.coeff /= len(funkslice);
  return [ outfunk ];

def linear_interpol (funkslice):
  """Reduction function to replace a set of funklets with piecewise linear interpolations. If the original
  N funklet c00 coefficients C0,C1,... are defined for domains [a0,b0] [a1,b1] ..., then the output N-1 funklets 
  will be defined on domains [a0,(a1+b1)/2],[(a1+b1)/2,(a2+b2)/2],..., and will correspond to linear slopes 
  passing through C0 at (a0+b0)/2, C1 at (a1+b1)/2, etc.
  """
  # transform only available with more than two funklets  
  if len(funkslice) < 2:
    return funkslice;
  if funkslice.rank > 1:
    raise TypeError,"linear interpolation only available for rank-1 slices";
  iaxis0 = funkslice.slice_iaxes[0]; 
  axis0 = funkslice.slice_axes[0]; 
  output = [];
  for ifunk0,funk0 in enumerate(funkslice[:-1]):
    # funklet for next domain
    funk1 = funkslice[ifunk0+1];
    dom0,dom1 = funk0.domain,funk1.domain;
    # centerpoint and c00 of ourselves
    x0 = sum(dom0[axis0])/2;
    c0 = funk0.coeff if isinstance(funk0.coeff,float) else funk0.coeff.ravel()[0];
    # centerpoint and c00 of next domain
    x1 = sum(dom1[axis0])/2;
    c1 = funk1.coeff if isinstance(funk1.coeff,float) else funk1.coeff.ravel()[0];
    # make output domain
    out_domain = copy.copy(dom0);
    # output domain boundaries are original subdomains' centers, with the exception of the first
    # and the last funklet, in which case the domain needs to extend to the edge of the first/last 
    # subdomain
    a0 = dom0[axis0][0] if funk0 is funkslice[0] else x0;
    a1 = dom1[axis0][1] if funk1 is funkslice[-1] else x1; 
    out_domain[axis0] = (a0,a1);
    # now make output polc
    output.append(meq.polc(coeff=[c0,c1-c0],domain=out_domain,offset=x0,scale=x1-x0,axis_index=iaxis0))
  return output;

def force_rank0 (funkslice):
  """Reduction function to reduce the polynomial rank of a set of funklets"""
  output = [];
  for funk in funkslice:
    c00 = funk.coeff if isinstance(funk.coeff,float) else funk.coeff.ravel()[0];
    output.append(meq.polc(coeff=c00,domain=funk.domain));
  return output;

_sub = dict([(a+b+c,b+c+':'+a) for a in 'ri' for b in 'xy' for c in 'xy' ]);

def rename (funkslice):
  fields = funkslice.name.rsplit(':',1);
  if len(fields) < 2: 
    return None;
  global _sub;
  rep = _sub.get(fields[-1]);
  if rep:
    fields[-1] = rep;
    return [':'.join(fields)] + list(funkslice);
  return None;
    


#!/usr/bin/python
# -*- coding: utf-8 -*-

#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os.path
import math
import sys
import re

import numpy
import numpy.ma
import Owlcat
import Owlcat.Console

from Owlcat.ParmTables import verbosity,dprint,dprintf

progress = Owlcat.Console.Reporter(timestamp=True);

class PlottableEntityProps (object):
  """A PlottableEntityProps objects describes the properties of a plottable entity.
  Instances of this will be e.g. RealValue, ComplexValue, JonesMatrix, etc."""
  def __init__ (self,name,plottables):
    """Creates a plottable entity description. 'name' is a short name, e.g. "real values".
    'plottables' is a list of rules for entities into real numbers than can be plotted, i.e.
    a list of ('rulename',func) tuples, where func transforms entities into float arrays."""
    self.name = name;
    self.plottables = plottables;
    self.plottables_dict = dict(plottables);

RealValue = PlottableEntityProps("real values",[("",lambda x:x)]);


ComplexValue = PlottableEntityProps("complex values",[
  ('ampl',   numpy.ma.absolute),
  ('phase',  lambda x:numpy.ma.masked_array(numpy.angle(x),x.mask,fill_value=x.fill_value)),
  ('real',   lambda x:x.real),
  ('imag',   lambda x:x.imag),
]);

JonesMatrix = PlottableEntityProps("Jones matrices",[
  ('norm',   lambda matrix:numpy.sqrt(sum(map(lambda x:x*numpy.ma.conjugate(x),matrix))))
]);

def _matrix22 (a11,*args):
  """helper function for aggregators below: makes 2x2 matrix. When called with 1 args, makes
  scalar-diagonal matrix, with 2 makes diagonal matrix, with 4 makes full matrix.""";
  if not args:
    return (a11,0,0,a11);
  elif len(args) == 1:
    return (a11,0,0,args[0]);
  elif len(args) == 3:
    return (a11,args[0],args[1],args[2]);
  raise TypeError,"incorrect number of arguments in call to _matrix22";

# 'Aggregators' is a list of rules for aggregating individual real parm coefficients into compound
# entities. E.g. ':r' and ':i' parms can be aggregated to form a complex number, then xx,xy,yx,yy
# complex values can be aggregated into a 2x2 matrix, etc.
# Each entry in the list has the form:
#   suffix_list,func,prop
# where 'suffix_list' gives the suffixes of the elements ('r','i'), 'func' is a callable for transforming
# the elements into an aggregate value, and 'prop' is a PlottableEntityProps object describing the aggregate.
#
Aggregators = [
  (('r','i'),(lambda r,i:r+1j*i),ComplexValue),
  (('r',),(lambda r:r),RealValue),
  (('ampl','phase'),(lambda ampl,phase:ampl*numpy.exp(1j*phase)),ComplexValue),
  (('ampl',),(lambda r:r),RealValue),
  (('xx','xy','yx','yy'),_matrix22,JonesMatrix),
  (('XX','XY','YX','YY'),_matrix22,JonesMatrix),
  (('rr','rl','lr','ll'),_matrix22,JonesMatrix),
  (('RR','RL','LR','LL'),_matrix22,JonesMatrix),
  (('xx','yy'),_matrix22,JonesMatrix),
  (('rr','ll'),_matrix22,JonesMatrix),
  (('XX','YY'),_matrix22,JonesMatrix),
  (('RR','LL'),_matrix22,JonesMatrix)
];

def _make_nested_func (agg_func,comp_func,n):
  """Helper method to make a nested aggregator function. Returns callable which takes any number of args a[],
  and returns agg_func(comp_func(a[0],...,a[n-1]),comp_func(a[n],...,a[2n-1]),...).
  If comp_func is None, returns simply arg_func."""
  if comp_func:
    return lambda *a:agg_func(*[comp_func(*a[i:i+n]) for i in range(0,len(a),n)]);
  else:
    return agg_func;

def makeParmEntityList (parmnames):
  """Processes list of parmnames and makes a list of plottable entities.
  Returns list of tuples:
    (entity_dict,ncomp,func,props), where:
  'desc' is a short description of this group of entities
  'entity_dict' is a dict of {name:components},
    'name' is an entity name (original parmname, or shortened aggregated name)
    'components' is a flat list of parmnames making up that aggregate
  'ncomp' is the number of components (each component list in the dict must have that many elements)
  'func' is a callable for transforming components into an aggregate value (None for individual parms)
  'props' is a PlottableEntityProps object.

  Each entry in the list corresponds to a group of entities that matched a particular aggregation rule.
  """;
  # start with one dict: that of all parms
  entity_list = [];
  new_entity_list = [ ( dict([ (name,[name]) for name in parmnames ]),1,lambda x:x,RealValue ) ];
  aggregated = set();
  # entities are moved from new_entity_list to entity_list as they're processed.
  while new_entity_list:
    entity_dict,ncomp,func,props = new_entity_list[0];
    # see if we have recipe for making compound objects based on last parmname qualifier
    nqlist = [ name.rsplit(':',1) for name in entity_dict.iterkeys() ];
    # set of basenames and set of all occurring qualifiers
    basenames = set([nq[0] for nq in nqlist]);
    quals = set([nq[1] for nq in nqlist]);
    for quallist,agg_func,agg_props in Aggregators:
      # look for an aggregator whose qualifiers are a subset of what we have
      if set(quallist) <= quals:
        new_entity_dict = {};
        # for each basename, attempt to find entities for every qualifier in the list
        for basename in basenames:
          fqnames = [ basename+':'+qq for qq in quallist ];
          components = [ name not in aggregated and entity_dict.get(name,None) for name in fqnames ];
          # If every qualified name gives a valid component list, make an aggregated entity
          # Components have to be transformed into a flat list.
          if all(components):
            new_entity_dict[basename] = reduce(lambda x,y:x+y,components);
            aggregated.update(fqnames);
        # if any new entities were aggregated, add to new list for firther reprocessing
        if new_entity_dict:
          new_entity_list.append((new_entity_dict,
                                  ncomp*len(quallist),
                                  _make_nested_func(agg_func,func,ncomp),
                                  agg_props));
    # move current entity to processed list
    entity_list.append(new_entity_list.pop(0));
  # flip list around, so latest match (presumably most complex entity) comes first
  entity_list.reverse();
  return entity_list;

def namesToSetNotation (namelist):
  # build set of possible values for each field
  fsets = {};
  for name in namelist:
    for i,f in enumerate(name.split(':')):
      fsets.setdefault(i,set()).add(f);
  # return compound string
  return ":".join([(fsets[i].pop() if len(fsets[i]) == 1 else "{%s}"%",".join(sorted(fsets[i])))
          for i in range(len(fsets))]);

if __name__ == "__main__":
  # setup some standard command-line option parsing
  #
  from optparse import OptionParser,OptionGroup
  parser = OptionParser(usage="""%prog: [options] <meptable> <plots...>""",
   description="Makes plots of parmtables. Each plot is specified as '[+]name/what'. "
    "E.g. 'G:*:xx/phase G:*:xx/ampl' plots all G XX phases in one plot, and amplitudes "
    "in a second plot (assuming your MEP table actually contains G XX solutions.) By contrast, "
    "'G:*:xx/phase +G:*:xx/ampl' will put phases and amplitudes into one plot, with amplitudes "
    "using the right Y-axis (NB: the \"+\" feature currently doen't work). "
    "A second form produces scatterplots: 'name/whatX-whatY'. For example, G:*:xx/real-imag "
    "will produce a complex scatterplot. "
    "Names may contain shell-style wildcards and {a,b,c} constructs; for more flexibility, "
    "use the -x option to interpret them as regexes instead. "
    "Use the -l (--list) option to see what can "
    "actually be plotted from your table. If you don't specify anything, the default is to plot "
    "whatever is the first group printed by -l."
  );
  parser.add_option("-x","--regex",dest="regex",action="store_true",
                    help="parse plot specifications as regexes instead of shell-style wildcards."
                   );
  parser.add_option("-l","--list",dest="list",action="store_true",
                    help="queries parmtable and lists what can be plotted");
#  parser.add_option("-a","--axis-select",dest="axis_select",metavar="axis=start[:end[:[+]step]]",type="string",action="append",
#                    help="Selects subset of data along an axis. E.g. 'time=0' selects the first timeslot, "
#                    "'freq=0:99' selects the first 100 frequencies, 'freq:0:99:2' selects every second frequency "
#                    "from the first 100, "
#                    "'time=::100' selects every 100th timeslot, 'time=::+100' averages every block of 100 "
#                    "timeslots. 'axis=all' or 'axis=::' selects everything. 'axis=avg' or 'axis=::+' selects and "
#                    "averages everything. The -a option may be given multiple times for different axes, or "
#                    "you can given one comma-separated list. The default for an axis that is not explicitly "
#                    "selected is 'all' for time and 'avg' for the rest.");
#  parser.add_option("-x","--xaxis",dest="xaxis",metavar="axis",type="string",
#                  help="Selects which data axis to use for the plotted X axis. Default is 'time'.");
  parser.add_option("--average-before",dest="average_before",action="store_true",
                    help="apply averaging to parameters BEFORE conversion to plottable values.");
  parser.add_option("--average-after",dest="average_after",action="store_true",
                    help="apply averaging to plottable values, AFTER conversion from parameters (default).");
  parser.add_option("--offset",type="float",
                    help="use a fixed offset between plot tracks. Default is automatically chosen.");
  parser.add_option("--min",type="float",default=None,
                    help="filter out values <MIN");
  parser.add_option("--max",type="float",default=None,
                    help="filter out values >MAX");

  group = OptionGroup(parser,"Plot label options");
  group.add_option("--label-mean",dest="label_mean",action="store_true",
                    help="include mean value in plot labels");
  group.add_option("--no-label-mean",dest="label_mean",action="store_false",
                    help="do not include mean value in plot labels");
  group.add_option("--label-stddev",dest="label_stddev",action="store_true",
                    help="include standard deviation in plot labels");
  group.add_option("--no-label-stddev",dest="label_stddev",action="store_false",
                    help="do not include standard deviation in plot labels");
  parser.add_option_group(group);

  group = OptionGroup(parser,"Output options");
  group.add_option("-o","--output",dest="output",type="string",
                    help="name of output file (e.g. 'foo.ps' or 'bar.png'). If not specified, plot will be displayed in a window.");
  group.add_option("--dpi",dest="resolution",type="int",metavar="DPI",
                    help="plot resolution for .ps/.png output (default is %default)");
  group.add_option("--plots-per-page",dest="ppp",metavar="N",type="int",
                    help="max plots per page. Default is 120.");
  group.add_option("--title",dest="title",type="string",
                    help="plot title (default is '<parmtab> <desc>')");
  group.add_option("-v","--verbose",dest="verbose",type="int",
                    help="set verbosity level for debugging messages");
  parser.add_option_group(group);

  parser.set_defaults(output="",xaxis="time",resolution=300,ppp=120,
    label_mean=True,label_stddev=True,offset=None,
  );

  (options,args) = parser.parse_args();
  if not args:
    parser.error("MEP table not specified. Use '-h' for help.");
  tabname = args[0];
  plotspecs = args[1:];

  verbosity.set_verbose(options.verbose);

  # read table, figure out what's in it
  progress("Reading MEP table %s, please wait..."%tabname);
  from Owlcat import ParmTables
  from Timba import mequtils
  pt = ParmTables.open(tabname);

  name_list = pt.parmtable().name_list();
  entity_list = makeParmEntityList(name_list);

  if options.list:
    print "\nTable contains %d funklets. We can plot the following things:\n"%len(name_list);
    for i,(entity_dict,ncomp,func,props) in enumerate(entity_list):
      names = namesToSetNotation(entity_dict.iterkeys());
      plts = ",".join([ p[0] for p in props.plottables ]);
      plts = plts and ("/{%s}"%plts if len(props.plottables)>1 else "/%s"%plts);
      print "  %s: %s%s%s"%(props.name,names,plts,("" if i else " (plotted by default)"));
    print "\nValid axes are:\n";
    cells = pt.envelope_cells();
    for iaxis in range(mequtils.max_axis):
      axis = mequtils.get_axis_id(iaxis);
      grid = cells.grid.get(axis);
      if grid is not None:
        print "  %s: %d points from %g to %g"%(axis,len(grid),grid[0],grid[-1]);
    print "";
    sys.exit(0);

  #
  # parse plot specifications
  #
  import fnmatch
  import re

  # This will contain a definitive list of all plots. Each plot is specified as a
  # [description,entity_list,entity_list2,scatterplot]. entity_list2 is valid for dual plots,
  # and is None for a single plot.
  # Scatterplot is None for normal plots, and a 2-tuple of axis labels, for scatter plots.
  Plots = [];

  # go through plotlist and find matching entities. regexes must match field-by-field
  if plotspecs:
    for plotspec in plotspecs:
      # '+' sign specifies second plot on previous page
      secondplot = (plotspec[0] == '+');
      if secondplot:
        if not Plots:
          parser.error("Can't use '+' in '%s': no preceding plot."%plotspec);
        if Plots[-1][2] is not None:
          parser.error("Can't use '+' in '%s': preceding plot is already dual."%plotspec);
        if Plots[-1][3]:
          parser.error("Can't use '+' in '%s': preceding plot is a scatterplot."%plotspec);
        plotspec = plotspec[1:];
      # split into name/what components
      names,what = plotspec.rsplit("/",1) if "/" in plotspec else (plotspec,None);
      # what may be an "X-Y" combo (for scatterplots)
      # in this case scatterplot is set to "X","Y", else to None.
      scatterplot = what.split("-",1) if what else [];
      if len(scatterplot) != 2:
        scatterplot = None;
      else:
        if secondplot:
          parser.error("Can't use '+' in '%s': scatterplots cannot be dual."%plotspec);
      # split name into fields at each ":", and translate into regexes
      if options.regex:
        regexes = map(re.compile,names.split(':'));
      else:
        # glob mode: translate the {a,b,c} case to a regex, and use
        # fnmatch.translate() for the rest      else
        regexes = [ re.compile( "(%s)"%('|'.join(ss[1:-1].split(","))) if (ss and ss[0]=='{' and ss[-1]=="}")
                                else fnmatch.translate(ss) ) for ss in names.split(":") ];
      # build list of what to plot, as (name,components,func,plotfunc,plotfunc2) tuples
      # plotfunc2 is None for nroaml plots, or a plotfunc for scatterplots
      plot_entlist = [];
      for entity_dict,ncomp,func,props in entity_list:
        for name,comps in sorted(entity_dict.iteritems()):
          # name must have same number of fields as the original regex, and all must match
          nf = name.split(':');
          if len(nf) == len(regexes) and all([r.match(f) for r,f in zip(regexes,nf)]):
            # check that 'what' is a plottable
            if what:
              if scatterplot:
                plotfuncs  = [ props.plottables_dict.get(wh) for wh in scatterplot ];
                plotfunc,plotfunc2 = plotfuncs;
              else:
                plotfuncs = [ props.plottables_dict.get(what) ];
                plotfunc,plotfunc2 = plotfuncs[0],None;
              if not all(plotfuncs):
                print "%s/%s cannot be plotted. Use the -l option to see what can."%(name,what);
                sys.exit(2);
            # if 'what' not given, use first item in list
            else:
              what,plotfunc = props.plottables[0];
              plotfunc2 = None;
            # add to list
            # print "Plotting %s %s '%s' %s %s"%(name,comps,what,func,plotfunc);
            plot_entlist.append((name,comps,func,plotfunc,plotfunc2));
      # have we found something to plot?
      if plot_entlist:
        if secondplot:
          Plots[-1][2] = plot_entlist;
        else:
          Plots.append([plotspec,plot_entlist,None,scatterplot]);
        progress("Plot %d: %s%s%s"%(len(Plots),namesToSetNotation([e[0] for e in plot_entlist]),
                    "/"+what if what else ""," (second axis)" if secondplot else ""));
      else:
        print "Nothing found to match '%s', ignoring."%plotspec;
  # if no plot options supplied, default is to take first group of plottables.
  else:
    entity_dict,ncomp,func,props = entity_list[0];
    what,plotfunc = props.plottables[0];
    progress("Plot 1: %s%s"%(namesToSetNotation(entity_dict.iterkeys()),"/"+what if what else ""));
    Plots.append(["",[(name,comps,func,plotfunc,None) for name,comps in sorted(entity_dict.iteritems())],None,False]);

  if not Plots:
    print "Found nothing to plot. Use the -l option to see what may be plotted.";
    sys.exit(0);

  # set non-interactive backend, if -o is in effect
  import matplotlib
  if options.output:
    matplotlib.use('agg');
  else:
    matplotlib.use('qt4agg');

  # ok, make the plots
  from Owlcat.Plotting import PlotCollection,ScatterPlot

  # make a page counter, if multiple pages are expected
  if len(Plots) > 1 or any([ len(entlist) > options.ppp for (desc,entlist,entlist2,scatterplot) in Plots if not scatterplot ]):
    ipage = 0;
  else:
    ipage = None;

  for iplot,(desc,entlist,entlist2,scatterplot) in enumerate(Plots):
    # setup title
    title = options.title or "%s: %s"%(tabname,desc);

    # make X-Y scatterplot
    if scatterplot:
      plot = ScatterPlot(options);
      plot.desc = desc;
      progress("Starting plot %d of %d, %d data points"%(iplot+1,len(Plots),len(entlist)));
      # loop over all plottables and compute data for each
      for name,comps,func,plotfunc,plotfunc2 in entlist:
        progress.overprint("Getting data for %s"%name);
        # get per-component arrays from ParmTable
        arrays = [ pt.funkset(comp).array() for comp in comps ];
        if options.average_before:
          arrays = map(numpy.ma.mean,arrays);
        dd = func(*arrays);
        x,y = plotfunc(dd),plotfunc2(dd);
        if options.average_after:
          x,y = map(numpy.ma.mean,(x,y));
        # add point
        plot.add_point(name,x,y);

      # make filename, if saving file
      savefile = options.output;
      if ipage is not None:
        if savefile:
          basename,ext = os.path.splitext(savefile);
          savefile = "%s.%d%s"%(basename,ipage,ext);
          progress("Plot will be written to file %s\n"%savefile);
        ipage += 1;
      progress("Generating plot");
      plot.make_figure(save=savefile,suptitle=title,dpi=options.resolution,xlabel=scatterplot[0],ylabel=scatterplot[1]);

    # normal "collections" plot
    else:
      # figure out label format
      labels = ["%(name)s"];
      if options.label_mean:
        labels.append("mean=%(mean).3g");
      if options.label_stddev:
        labels.append("std=%(stddev).3g");
      label_format = " ".join(labels);
      # make collection object
      coll = PlotCollection(options);
      if options.offset:
        coll.offset = options.offset;
      coll.desc = desc;
      progress("Starting plot %d of %d, %d data tracks"%(iplot+1,len(Plots),len(entlist)));
      # loop over all plottables and compute data for each
      for name,comps,func,plotfunc,plotfunc2 in entlist:
        progress.overprint("Getting data for %s"%name);
        # get per-component arrays from ParmTable
        arrays = [ pt.funkset(comp).array() for comp in comps ];
        if options.average_before:
          arrays = [ arr.mean(1) for arr in arrays ];
        # apply functions to get plottable values
        dd = plotfunc(func(*arrays));
        # apply after-averaging, if needed
        if not options.average_before:
          dd = dd.mean(1);
        mask = None;
        if options.min is not None:
          mask = dd < options.min;
        if options.max is not None:
          mask = mask&(dd > options.max) if mask is not None else (dd > options.max);
        if mask is not None:
          dd = numpy.ma.masked_array(dd,mask);
        # put together label
        labelattrs = dict(name=name,mean=dd.mean(),stddev=dd.std());
        # add track
        coll.add_track(name,dd,label=label_format%labelattrs);

      for np0 in range(0,len(entlist),options.ppp):
        np1 = min(np0+options.ppp,len(entlist));
        names = [ ent[0] for ent in entlist[np0:np1] ];
        # make filename, if saving file
        savefile = options.output;
        if ipage is not None:
          if savefile:
            basename,ext = os.path.splitext(savefile);
            savefile = "%s.%d%s"%(basename,ipage,ext);
            progress("Plot will be written to file %s\n"%savefile);
          ipage += 1;
        # dual plot?
        dual = (len(names) > options.ppp/2);
        progress("Generating plot");
        coll.make_figure(names,save=savefile,suptitle=title,dpi=options.resolution,dual=dual);

  if not options.output:
    from pylab import plt
    plt.show();





#!/usr/bin/python
# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import os.path
import re
import time

from Owlcat import table,tablecopy,tableexists,tabledelete

if __name__ == "__main__":

  # setup some standard command-line option parsing
  #
  from optparse import OptionParser
  parser = OptionParser(usage="""%prog: [options] MS""",
    description="Splits MS into one or more output MSs by DATA_DESC_ID (which typically corresponds "
    "to spectral window.)");
  parser.add_option("-D","--ddid",type="int",
                    help="specific DATA_DESC_ID to extract. If not given, then all DDIDs are extracted, each into a separate MS.");
  parser.add_option("-o","--output",type="string",
                    help="Name of output MS. Use '%(ms)s' to insert basename of input MS. Use '%(ddid)d' to insert DDID number. Use '%(msext)s' to insert extension of input MS. Default is %default");
  parser.add_option("-f","--force",dest="force",action="store_true",
                    help="overwrites output MS if it already exists");

  parser.set_defaults(ddid=-1,output="%(ms)s_spw%(ddid)d%(msext)s");

  (options,msname) = parser.parse_args();

  if len(msname) != 1 :
    parser.error("Incorrect number of arguments. Use '-h' for help.");
  
  # open input MS
  msname = msname[0];
  print "Reading input MS %s"%msname;
  if not tableexists(msname):
    parser.error("MS %s not found."%msname);
  ms = table(msname);
  # check DDID option
  num_ddids = table(ms.getkeyword("DATA_DESCRIPTION")).nrows();
  print "MS contains %d DATA_DESC_IDs"%num_ddids;
  if options.ddid < 0:
    ddids = range(num_ddids);
  elif options.ddid >= num_ddids:
    parser.error("DATA_DESC_ID %d is out of range"%options.ddid);
  else:
    ddids = [ options.ddid ];

  # setup outputs
  msname,msext = os.path.splitext(os.path.basename(os.path.normpath(msname)));

  for ddid in ddids:
    msout = options.output%dict(ms=msname,msext=msext,ddid=ddid);
    print "Extracting DATA_DESC_ID %d into MS %s"%(ddid,msout);
    # delete if exists
    if tableexists(msout):
      if not options.force:
        parser.error("Output MS %s already exists. Use the -f switch to overwrite."%msout);
      print "Deleting existing copy of %s"%msout;
      tabledelete(msout);
    # extract
    ms1 = ms.query("DATA_DESC_ID==%d"%ddid);
    ms1.copy(msout,deep=True);
    ms1.close();




#!/usr/bin/python
# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

if __name__ == "__main__":

  # setup some standard command-line option parsing
  #
  from optparse import OptionParser,OptionGroup
  from Owlcat.Plotting import MultigridPlot,SkyPlot,PLOT_SINGLE,PLOT_MULTI,PLOT_ERRORBARS

  parser = OptionParser(usage="""%prog: [plots & options] parmtables""",
      description="""Makes various plots of poitning error solutions.""");

  parser.add_option("-l","--list",action="store_true",
                    help="lists stuff found in MEP tables, then exits");
  parser.add_option("-c","--cache",metavar="FILENAME",type="string",
                    help="cache parms to file, which can be fed to plot-de-solutions script");
  parser.add_option("-n","--nominals",metavar="FILENAME",type="string",
                    help="includes nominal offsets on plot (supply filename)");
  parser.add_option("--nominal-circle",metavar="mDEG",type="float",default=10,
                    help="size of circle around nominal offsets");
  parser.add_option("--circle-radius",metavar="mDEG",type="float",default=0,
                    help="set a fixed size for the circle pointing plot");
  parser.add_option("--from",metavar="TIMESLOT",type="int",default=0,
                    help="extract solutions starting from given timeslot");
  parser.add_option("--to",metavar="TIMESLOT",type="int",default=-1,
                    help="extract solutions until the given timeslot");
  parser.add_option("--antennas",type="string",
                    help="antenna subset, as a comma-separated list of names. Simple wildcards are allowed.");
  parser.add_option("--pt-max",metavar="VALUE",type="float",default=0,
                    help="set fixed plot limits on pointing error vs. time plot");
  parser.add_option("--ft",action="store_true",
                    help="generate plots of pointing offset Fourier components");
  parser.add_option("--ft-max",metavar="VALUE",type="float",default=0,
                    help="set fixed plot limits on Fourier component plot");
  parser.add_option("--wind",metavar="FILE",type="string",
                    help="loads wind data from pickle file, adds wind plots");

  plotgroup = OptionGroup(parser,"Plotting options");
  outputgroup = OptionGroup(parser,"Output options");
  MultigridPlot.init_options(plotgroup,outputgroup);
  SkyPlot.init_options(plotgroup,outputgroup);

  parser.add_option_group(plotgroup);
  parser.add_option_group(outputgroup);

  (options,args) = parser.parse_args();

  if not args:
    parser.error("No parmtables specified.");

  import os.path
  import os
  import sys

  import numpy
  from Owlcat.ParmTables import ParmTab
  import math
  DEG = math.pi/180;
  ARCMIN = math.pi/(180*60);

  SPWS = range(len(args));

  # set of all sources, antennas and correlations
  ANTS = set();

  # complex array of dEs per each src,antenna,corr tuple
  des = {};

  # scan funklet names to build up sets of keys
  oldtable = False;
  pt = ParmTab(args[0]);
  for name in pt.funklet_names():
    if name.startswith("E::dlm::dl"):
      oldtable = True;
    if name.startswith("E::dlm::dl:") or name.startswith("E::dl:"):
      fields = name.split(':');
      ANTS.add(fields[-1]);
  ts_slice = slice(getattr(options,'from'),options.to if options.to >=0 else None);
  NTIMES = len(pt.funkset(pt.funklet_names()[0]).get_slice()[ts_slice]);
  print "%d timeslots found in table %s"%(NTIMES,args[0]);

  ANTS = sorted(ANTS);
  if not ANTS:
    print "No pointing offset solutions found in MEP table %s."%args[0];
    sys.exit(1);
    
  if options.list:
    print "MEP table %s contains pointing offsets for"%args[0];
    print "  %d antennas: %s"%(len(ANTS)," ".join(ANTS));
    
  # source subset selection
  import fnmatch
  if options.antennas:
    ants = set();
    for a in options.antennas.split(","):
      subset = fnmatch.filter(ANTS,a);
      if subset:
        ants.update(subset);
      else:
        print "WARNING: \"%s\" does not match any antenna names in MEP table %s."%(a,args[0]);
    ANTS = sorted(ants);
    if not ANTS:
      print "No antennas were selected, check your --antennas option.";
      sys.exit(1);
      
  print "Using %d antennas: %s"%(len(ANTS)," ".join(ANTS));

  interval = None;

  def c00 (funklet):
    global interval;
    if interval is None:
      interval = funklet.domain.time[1] - funklet.domain.time[0];
    if numpy.isscalar(funklet.coeff):
      return funklet.coeff;
    else:
      return funklet.coeff.ravel()[0];

  # dl,dm is a 2 x NSPW x  NANT x NTIME array of poitning offsets
  dlm = numpy.zeros((2,len(SPWS),len(ANTS),NTIMES),dtype=float);

  # bsz is a 2 x 2 x NSPW x  NANT x NTIME array of beam sizes
  # first index is x/y, second is l/m
  bsz = numpy.zeros((2,2,len(SPWS),len(ANTS),NTIMES),dtype=float);
  beam_sizes = 0;
  
  for spw,tabname in enumerate(args):
    print "Reading",tabname;
    pt = ParmTab(tabname);
    for i,ant in enumerate(ANTS):
      # fill times
      # fill dlm
      if oldtable:
        fsl = pt.funkset('E::dlm::dl:%s'%ant).get_slice()[ts_slice];
        fsm = pt.funkset('E::dlm::dm:%s'%ant).get_slice()[ts_slice];
      else:
        fsl = pt.funkset('E::dl:%s'%ant).get_slice()[ts_slice];
        fsm = pt.funkset('E::dm:%s'%ant).get_slice()[ts_slice];
      if len(fsl) != len(fsm) or len(fsl) != NTIMES:
        print "Error: table contains %d funklets for dl and %d for dm; %d expected"%(len(fsl),len(fsm),NTIMES);
        sys.exit(1);
      if not i:
        t0 = numpy.array([ funklet.domain.time[0] for funklet in fsl ])
        t1 = numpy.array([ funklet.domain.time[1] for funklet in fsl ]);
      dlm[0,spw,i,:] = map(c00,fsl);
      dlm[1,spw,i,:] = map(c00,fsm);
      # fill beam sizes
      if 'E::beamshape:%s'%ant in pt.funklet_names():
        beam_sizes = 1;
        fs = pt.funkset('E::beamshape:%s'%ant).get_slice()[ts_slice];
        bsz[0,0,spw,i,:] = map(c00,fs);
      elif 'E::beamshape:xy:lm:%s'%ant in pt.funklet_names():
        beam_sizes = 4;
        for ixy,xy in enumerate("xy"):
          for ilm,lm in enumerate("lm"):
            fs = pt.funkset('E::beamshape:%s:%s:%s'%(ant,xy,lm)).get_slice()[ts_slice];
            bsz[ixy,ilm,spw,i,:] = map(c00,fs);
  
  interval = round((interval or 60)/60);
  print "Solution interval is",interval,"minutes; total time",(interval*NTIMES)/60,"hours";

  # write cache
  if options.cache:
    import cPickle
    cachefile = options.cache+'.cache';
    cPickle.dump((dlm,bsz,beam_sizes),file(cachefile,'w'));
    print "Cached all structures to file",cachefile;
    
  if options.wind:
    import cPickle
    wind_time,wind_dir,wind_speed = cPickle.load(file(options.wind));
    print "Loaded %d wind samples from %s"%(len(wind_time),options.wind);
    # get wind data for each interval
    wsp_mean = numpy.zeros(NTIMES,float);
    wsp_std = numpy.zeros(NTIMES,float);
    for itime in range(NTIMES):
      i0 = wind_time.searchsorted(t0[itime]);
      i1 = wind_time.searchsorted(t1[itime]);
      if i1 - i0 <= 0:
        print "No wind data for interval %d, cannot plot wind"%itime;
        sys.exit(1);
      wsp_mean[itime] = wind_speed[i0:i1].mean();
      wsp_std[itime] = wind_speed[i0:i1].std();
    

  # convert dlm to millidegrees
  dlm0 = dlm.copy();
  dlm *= 180*1000/math.pi;

  # take mean and std along freq axis
  # these are now 2 x NANT x NTIME arrays
  dlm_mean = dlm.mean(1);
  dlm_std  = dlm.std(1);
  bsz_mean = bsz.mean(2);
  bsz_std  = bsz.std(2);
  # and along time axis
  # these are now 2 x NSPW x NANT
  dlm_fmean = dlm.mean(3);
  dlm_fstd  = dlm.std(3);
  bsz_fmean = bsz.mean(4);
  bsz_fstd  = bsz.std(4);

  print "Read %d parmtables"%len(args);

  from Owlcat.Plotting import MultigridPlot,PLOT_SINGLE,PLOT_MULTI,PLOT_ERRORBARS,PLOT_BARPLOT

  # initialize plot object
  figplot = MultigridPlot(options);
  make_figure = figplot.make_figure;

  skyplot = SkyPlot(options);
  make_skymap = skyplot.make_figure;


  funcs = [
    lambda iant:(dlm_mean[0,iant,:],dlm_std[0,iant,:]),
    lambda iant:("mean %.2f"%numpy.array([dlm_mean[0,iant,:].mean()]),
                 "+/- %.2f"%numpy.array([dlm_mean[0,iant,:].std()])),
    lambda iant:(dlm_mean[1,iant,:],dlm_std[1,iant,:]),
    lambda iant:( "mean %.2f"%numpy.array([dlm_mean[1,iant,:].mean()]),
                  "+/- %.2f"%numpy.array([dlm_mean[1,iant,:].std()])),
    lambda iant:(dlm_fmean[0,:,iant],dlm_fstd[0,:,iant]),
    lambda iant:(dlm_fmean[1,:,iant],dlm_fstd[1,:,iant])
  ];
  
  labels = [ "dl","","dm","","dl, freq","dm, freq" ];
  make_figure(enumerate(labels),enumerate(ANTS),
        lambda i,iant:funcs[i](iant),
      hline=0,ylock=(-options.pt_max,options.pt_max) if options.pt_max else True,figsize=(290,150),mode=PLOT_ERRORBARS,
      suptitle="Pointing offset mean & stddev across all bands (top two plots) and times (bottom two plots), mdeg",
      save="Epnt_mean");
  
  if options.wind:
    # work out wind data for each solve interval
    funcs = [
      lambda iant:(dlm_mean[0,iant,:],dlm_std[0,iant,:]),
      lambda iant:(dlm_mean[1,iant,:],dlm_std[1,iant,:]),
      lambda iant:(wsp_mean,wsp_std),
    ];
    
    labels = [ "dl","dm","wind, m/s" ];
    make_figure(enumerate(labels),enumerate(ANTS),
          lambda i,iant:funcs[i](iant),
        hline=0,ylock="row",figsize=(290,75),mode=PLOT_ERRORBARS,
        suptitle="Pointing offset mean & stddev, versus wind speed",
        save="Epnt_wind");
    


  for iant,ant in enumerate(ANTS):
    print "mean offset %s: %6.2f %6.2f"%(ant,dlm_mean[0,iant,:].mean(),dlm_mean[1,iant,:].mean());

  if options.ft:
    import numpy.fft
    # fft scaling is 1/NTIMES, then later we multiply the non-0th-order terms by 2 since each fourier component's
    # power is split between itself and the conjugate 
    ft_slice = slice(0,NTIMES/2+1);
    dlm_fft = abs(numpy.fft.fft(dlm_mean))[:,:,ft_slice]/NTIMES;
    dlm_fft[:,:,1:] *= 2;
    periods = 1/abs(numpy.fft.fftfreq(NTIMES,interval)[1:]);
    dlm_fftmax = dlm_fft[:,:,1:].max(2);
    dlm_fftper = numpy.zeros((2,len(ANTS)));
    for i in 0,1:
      for iant in range(len(ANTS)):
        dlm_fftper[i,iant] = periods[numpy.where(dlm_fft[i,iant,1:] == dlm_fftmax[i,iant])]
    funcs = [
      lambda iant:dlm_fft[0,iant,:],
      lambda iant:("offset %.1f"%dlm_fft[0,iant,0],"~max %.1f"%dlm_fftmax[0,iant],"@%d min"%dlm_fftper[0,iant]),
      lambda iant:dlm_fft[1,iant,:],
      lambda iant:("offset %.1f"%dlm_fft[1,iant,0],"~max %.1f"%dlm_fftmax[1,iant],"@%d min"%dlm_fftper[1,iant]),
    ];
    labels = [ "FT(dl)","","FT(dm)","" ];
    make_figure(enumerate(labels),enumerate(ANTS),
          lambda i,iant:funcs[i](iant),
        hline=0,ylock=(0,options.ft_max or dlm_fft.max()),figsize=(290,25*len(labels)),mode=PLOT_BARPLOT,
        suptitle="Pointing offset Fourier components",
        save="Epnt_ft");


  if beam_sizes == 4:
    funcs = [];
    for i0 in range(2):
      for j0 in range(2):
        funcs += [
          lambda iant,i=i0,j=j0:(bsz_mean[i,j,iant,:],bsz_std[i,j,iant,:]),
          lambda iant,i=i0,j=j0:("mean %.2f"%numpy.array([bsz_mean[i,j,iant,:].mean()]),
                        "+/- %.2f"%numpy.array([bsz_mean[i,j,iant,:].std()]))
        ];

    make_figure(enumerate(("Lx","","Mx","","Ly","","My","")),enumerate(ANTS),
          lambda i,iant:funcs[i](iant),
        hline=1,ylock=True,figsize=(290,210),mode=PLOT_ERRORBARS,
        mean_format="%.4f",
        suptitle="Beam extent in L/M, for X and Y dipoles, mean over time",
        save="Eshape_mean");

    funcs = [];
    for i0 in range(2):
      for j0 in range(2):
        funcs += [
          lambda iant,i=i0,j=j0:(bsz_fmean[i,j,:,iant],bsz_fstd[i,j,:,iant]),
          lambda iant,i=i0,j=j0:("mean %.2f"%numpy.array([bsz_fmean[i,j,:,iant].mean()]),
                        "+/- %.2f"%numpy.array([bsz_fmean[i,j,:,iant].std()]))
        ];
    make_figure(enumerate(("Lx","","Mx","","Ly","","My","")),enumerate(ANTS),
          lambda i,iant:funcs[i](iant),
        mean_format="%.4f",
        hline=1,ylock=True,figsize=(290,210),mode=PLOT_ERRORBARS,
        suptitle="Beam extent in L/M, for X/Y dipoles, mean over frequency",
        save="Eshape_mean_fq");
  elif beam_sizes == 1:
    funcs = [
      lambda iant:(bsz_mean[0,0,iant,:],bsz_std[0,0,iant,:]),
      lambda iant:( "mean %.2f"%numpy.array([bsz_mean[0,0,iant,:].mean()]),
                    "+/- %.2f"%numpy.array([bsz_mean[0,0,iant,:].std()])),
      lambda iant:(bsz_fmean[0,0,:,iant],bsz_fstd[0,0,:,iant]),
    ];
    make_figure(enumerate(("size","","size fq")),enumerate(ANTS),
          lambda i,iant:funcs[i](iant),
        mean_format="%.4f",
        hline=1,ylock=True,figsize=(290,75),mode=PLOT_ERRORBARS,
        suptitle="Beam extent",
        save="Eshape");

  # make skymap with average pointings
  ll = [];
  mm = [];
  markers = [];

  # add nominal mispointings
  if options.nominals:
    exec(file(options.nominals));
  else:
    nominals = {};


  circle_radius = options.circle_radius/(1000/60.);
  nominal_radius = options.nominal_circle/(1000/60.);

  circlex1 = numpy.cos(numpy.arange(0,1.05,.05)*math.pi*2);
  circley1 = numpy.sin(numpy.arange(0,1.05,.05)*math.pi*2);
  circlex = circlex1*nominal_radius;
  circley = circley1*nominal_radius;

  ll.append(0);
  mm.append(0);
  markers.append(
    ("plot",(circlex,circley,":"),
        dict(color='blue'))
  );

  for iant,ant in enumerate(ANTS):
    dl = dlm0[0,:,iant,:];
    dm = dlm0[1,:,iant,:];
    dl_mean = dl.mean();
    dl_std = dl.std();
    dm_mean = dm.mean();
    dm_std = dm.std();
    color = "pink" if nominals else "blue";
    # plot nominal position, if available
    if ant in nominals:
      dl0,dm0 = nominals[ant];
      markers += [
        ("text",(dl0/ARCMIN,dm0/ARCMIN,ant),
            dict(color='blue',ha='center',va='center',size='large',weight='bold')),
        ("plot",(dl0/ARCMIN+circlex,dm0/ARCMIN+circley,":"),
            dict(color='blue')),
        ("plot",((dl0/ARCMIN,dl_mean/ARCMIN),(dm0/ARCMIN,dm_mean/ARCMIN),':'),
            dict(color='grey')),
        ];
      ll += [ dl0,dl0,dl0 ];
      mm += [ dm0,dl0,dl0 ];
      color = "red";
    # plot fitted position
    markers += [
      ("errorbar",(dl_mean/ARCMIN,dm_mean/ARCMIN,dm_std/ARCMIN,dl_std/ARCMIN),dict(color='#A0A0A0')),
      ("text",(dl_mean/ARCMIN,dm_mean/ARCMIN,"%s "%ant),
          dict(color=color,ha='center',va='center',size='large',weight='bold',bbox=dict(fc='white',ec='none' if beam_sizes else 'grey'))),
      ];
    ll += [ dl_mean,dl_mean ];
    mm += [ dm_mean,dm_mean ];
    # plot beam size
    if beam_sizes == 1:
      bsz = bsz_mean[0,0,iant,:].mean();
      bsz = min(1.2,max(bsz,.8));
      bsz = bsz - 1;
      radius0 = nominal_radius/4;
      radius = radius0 + bsz*5*radius0;
      markers += [ ("plot",
            ( dl_mean/ARCMIN+circlex1*radius0,
              dm_mean/ARCMIN+circley1*radius0,'-'),
          dict(color='grey',zorder=10,alpha=0.5)) ];
      markers += [ ("plot",
            ( dl_mean/ARCMIN+circlex1*radius,
              dm_mean/ARCMIN+circley1*radius,'-'),
          dict(color='green' if bsz >.01 else 'purple' if bsz <-.01 else 'grey',zorder=20,alpha=0.5)) ];
      ll += [ dl_mean,dl_mean ];
      mm += [ dm_mean,dm_mean ];

  make_skymap(numpy.array(ll),numpy.array(mm),markers,
    zero_lines=False,radius=circle_radius or None,
    suptitle="Fitted pointing offsets",save="Eplot");


  if options.output_type.upper() == "X11":
    from pylab import plt
    plt.show();



#!/usr/bin/python
# -*- coding: utf-8 -*-

#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os.path
import sys
import traceback
import gzip
import cPickle
import Owlcat

flagger = parser = ms = msname = None;

def error (message):
  print "%s: %s"%(os.path.basename(sys.argv[0]),message);
  sys.exit(1);

def get_ms ():
  global ms;
  global msname;
  if not ms:
    ms = Owlcat.table(msname);
  return ms;

if __name__ == "__main__":

  # setup some standard command-line option parsing
  #
  from optparse import OptionParser,OptionGroup
  parser = OptionParser(usage="""%prog: [actions] [options] MS COLUMN FILENAME[.gz]""",
      description="Exports MS column to an external file, which can be reloaded with import-ms-column."
  );
  parser.add_option("-r","--row-step",type="int",
                    help="how many rows to step over at a time, default is %default");
#  parser.add_option("-z","--gzip",action="store_true",
#                    help="compress data with gzip");
#  parser.add_option("-v","--verbose",metavar="LEVEL",type="int",
#                    help="verbosity level for messages. Higher is more verbose, default is 0.");
  parser.set_defaults(verbose=0,row_step=200000);

  # parse args
  (options,args) = parser.parse_args();
  if len(args) != 3:
    parser.error("Incorrect number of arguments. Use '-h' for help.");
  msname,colname,filename = args;

  try:
    gzf = gzip.GzipFile(filename,"w") if filename.endswith(".gz") else file(filename,"w");
    # write stuff
    ms = get_ms();
    print "Opened MS %s"%msname;
    if colname not in ms.colnames():
      error("Column %s not found"%colname);
    nrows = ms.nrows();
    print "Exporting %s to %s:"%(colname,filename);
    cPickle.dump((options.row_step,nrows),gzf);
    for r0 in range(0,nrows,options.row_step):
      print "%d/%d"%(r0,nrows);
      col = ms.getcol(colname,r0,options.row_step);
      cPickle.dump(col,gzf);
    ms.close();
    gzf.close();
  except:
    traceback.print_exc();
    error("Error exporting %s to %s"%(colname,filename));
  print "Column exported OK.";



# -*- coding: utf-8 -*-
import numpy
import numpy.ma
import math
import sys
import traceback
import os.path
import matplotlib

DEG = math.pi/180;
ARCMIN = math.pi/(180*60);


def init_pyplot (output_type="X11"):
  """Initializes the pyplot singleton. Inits the matplotlib back-end
  corresponding to the specified plot type""";
  if 'pyplot' not in globals():
    if output_type.upper() != "X11":
      matplotlib.use('agg');
    else:
      matplotlib.use('qt4agg');
    global pyplot;
    import matplotlib.pyplot as pyplot

def _int_or_0 (x):
  try:
    return int(x);
  except:
    return 0;
_version = map(_int_or_0,matplotlib.__version__.split('.'));
_version_major = (_version and _version[0]) or 0;
_version_minor = (len(_version)>1 and _version[1]) or 0;

class PlotCollection (object):
  """PlotCollection plots a collection of data tracks in one plot""";
  def __init__ (self,options):
    self.data = {};
    self.mean = {};
    self.stddev = {};
    self.label = {};
    self.track_counts = {};
    # inter-plot offset -- may be set from outside
    self.offset = 0;

  def num_tracks (self):
    return len(self.data);

  def add_track (self,key,data,mean=None,stddev=None,label=None,count=1):
    self.data[key] = data;
    self.mean[key] = mean if mean is not None else data.mean();
    self.stddev[key] = stddev if stddev is not None else data.std();
    self.label[key] = label if label is not None else str(key);
    self.track_counts[key] = count;

  def get_track_data (self,key):
    return self.data.get(key,None);

  def get_track_count (self,key):
    return self.track_counts.get(key,0);

  # function to make single-page plot
  def make_figure (self,keylist=None,suptitle=None,save=None,
                   dual=True,offset_std=10,xgrid=False,ygrid=False,
                   figsize=(210,290),dpi=100,papertype='a4',landscape=False):
    import matplotlib.pyplot as pyplot
    # setup sizes
    keylist = keylist or sorted(self.data.keys());
    # compute inter-plot offset, based on the median stddev
    offset = self.offset;
    if not offset:
      stddevs = sorted(self.stddev.itervalues());
      if stddevs:
        offset = self.offset = offset_std*stddevs[len(stddevs)/2];
    # if still not set, force to 1
    if not offset:
      offset = self.offset = 1;
    # partition keys into two sets, if asked to
    if dual:
      # split ifrs into two sets (if number is odd, make sure first subset has the extra 1)
      n = len(keylist);
      n2 = len(keylist)/2;
      if n%2:
        n2 += 1;
      keysets = [ keylist[0:n2],keylist[n2:] ];
    else:
      keysets = [ keylist ];
    # create Figure object of given size and resolution
    figsize_in = (figsize[0]/25.4,figsize[1]/25.4);
    fig = pyplot.figure(figsize=figsize_in,dpi=100);
    # margin sizes. We want to keep them fixed in absolute terms,
    # regardless of plot size. The relative numbers here are good for a 210x290 plot, so
    # we rescale them accordingly.
    mleft   = 0.05 * 210./figsize[0];
    mbottom = 0.03 * 290./figsize[1];
    mright  = 0.01 * 210./figsize[0];
    mtop    = 0.03 * 290./figsize[1];
    ytitle  = 1 - 0.01 * 290./figsize[1];
    width   = ( 1 - mleft - mright );
    height  = ( 1 - mtop - mbottom );
    # now plot all tracks
    for iplot,keys in enumerate(keysets):
      if dual:
        plt = fig.add_subplot(1,2,iplot+1);
      else:
        plt = fig.add_axes([mleft,mbottom,width,height]);
      nx = max([len(self.data.get(key,[])) for key in keys]);
      # first and last key
      key0 = keys[0];
      key1 = keys[-1];
      mindata,maxdata = 1e+99,-1e+99;
      # make plots for all ifrs
      firstkey = None;
      for key in keys:
        data = self.data.get(key);
        if data is None or data.mask.all():
          continue;
        else:
          # print data.min(),data.max();
          # first plot goes at its natural coordinates.
          if firstkey is None:
            firstkey = key;
            y0 = 0;
            y0text = self.mean[key] - offset/2;
            mindata = data.min() - offset/2;
#            mindata = y0text = self.mean[key] - offset/2;
            dd = data;
          # subsequent plots offset accordingly
          else:
            y0 += offset;
            dd = data + y0;
#          maxdata = self.mean[key] + y0 + offset/2;
          mindata = min(mindata,data.min() + y0 - offset/10);
          maxdata = max(maxdata,self.mean[key] + y0 + offset);
          # plot data
          try:
            if len(dd) == 1:
              dd = numpy.array([dd[0],dd[0]]);
            plt.plot(dd);
            if ygrid:
              plt.axhline(y0,color='0.8',zorder=-10);
            #print dd;
          except:
            traceback.print_exc();
            print "Error plotting data for",key;
            continue;
          if self.label[key]:
            # figure out where to place text label
            ytext = dd[:nx/10].mean();
            if numpy.isnan(ytext):
              ytext = dd.mean();
            # do not put too close to previous label
            y0text = max(y0text+offset/4,ytext);
            # annotate() seems to have a bug in matplotlib 0.98, so use text() instead
            if _version_major > 0 or _version_minor > 98:
              plt.annotate(self.label[key],xy=(nx/100,ytext),xytext=(nx/100,y0text),
                    size=5,horizontalalignment='left',verticalalignment='center',
                    bbox=dict(facecolor='white',edgecolor='none',alpha=0.6),
                    arrowprops=dict(fc='none',width=0,headwidth=0,ec="0.8",alpha=0.6));
            else:
              plt.text(nx/100,ytext,self.label[key],
                    size=5,horizontalalignment='left',verticalalignment='center',
                    bbox=dict(facecolor='white',edgecolor='none',alpha=0.6));

      # set plot limits. First panel is determined by data. Scale of second panel is fixed to first.
      plt.set_xbound(0,nx);
      if iplot == 0:
        ylim = mindata,maxdata;
        plt.set_ybound(*ylim);
        for lab in plt.get_yticklabels():
          lab.set_fontsize(5);
      else:
        for lab in plt.get_yticklabels():
          lab.set_fontsize(0);
        # plt.set_yticklabels([]);
        plt.set_ybound(*ylim);
      for lab in plt.get_xticklabels():
        lab.set_fontsize(5);
      # set x grid
      if _version_major > 0 or _version_minor > 98:
        plt.minorticks_on();
      if xgrid and xgrid != [0]:
        if isinstance(xgrid,(list,tuple)):
          if len(xgrid)>0:
            plt.set_xticks(range(0,nx,xgrid[0]));
          if len(xgrid)>1:
            plt.set_xticks(range(0,nx,xgrid[1]),True);
        for x in plt.get_xticks():
          plt.axvline(x,color='0.7',ls='-',lw=1,zorder=-10);
        for x in plt.get_xticks(True):
          plt.axvline(x,color='0.9',ls='-',lw=1,zorder=-10);
        plt.set_xlim(0,nx);
    fig.subplots_adjust(left=mleft,right=1-mright,top=1-mtop,bottom=mbottom,wspace=0.01);
    # plot title if asked to
    if suptitle:
      fig.suptitle(suptitle,y=ytitle,size=8);
    if save:
      fig.savefig(save,papertype=papertype,dpi=dpi,
                  orientation='portrait' if not landscape else 'landscape');
      print "===> Wrote",save;
    return fig;

class PlotCollectionSep (PlotCollection):
  """PlotCollectionSep is a PlotCollection that separates data tracks into individual plots,
  and optionally groups them by key""";
  def __init__ (self,options):
    PlotCollection.__init__(self,options);
    self.groups = {};

  def num_tracks (self):
    return len(self.data);

  def add_track (self,key,data,mean=None,stddev=None,label=None,count=1,group=None):
    self.groups.setdefault(group,set()).add(key);
    PlotCollection.add_track(self,key,data,mean=mean,stddev=stddev,label=label,count=count);

  # function to make single-page plot
  def make_figure (self,grouplist=None,suptitle=None,save=None,
                   dual=True,xgrid=False,ygrid=False,
                   figsize=(210,290),dpi=100,papertype='a4',landscape=False):
    import matplotlib.pyplot as pyplot
    # setup sizes
    grouplist = grouplist or sorted(self.groups.keys());
    # partition keys into two sets, if asked to
    if dual:
      # split ifrs into two sets (if number is odd, make sure first subset has the extra 1)
      n = len(grouplist);
      nrow = len(grouplist)/2;
      if n%2:
        nrow += 1;
      # swap things around, so plots start at the bottom
      plotnums = numpy.arange(1,nrow*2+1).reshape(nrow,2)[-1::-1].transpose().ravel()
    else:
      nrow = len(grouplist);
      plotnums = numpy.arange(nrow,0,-1);
    # create Figure object of given size and resolution
    figsize_in = (figsize[0]/25.4,figsize[1]/25.4);
    fig = pyplot.figure(figsize=figsize_in,dpi=100);
    # margin sizes. We want to keep them fixed in absolute terms,
    # regardless of plot size. The relative numbers here are good for a 210x290 plot, so
    # we rescale them accordingly.
    mleft   = 0.05 * 210./figsize[0];
    mbottom = 0.03 * 290./figsize[1];
    mright  = 0.01 * 210./figsize[0];
    mtop    = 0.03 * 290./figsize[1];
    ytitle  = 1 - 0.01 * 290./figsize[1];
    width   = ( 1 - mleft - mright );
    height  = ( 1 - mtop - mbottom );
    colors = ( "blue","green","red","cyan","purple","magenta","black","orange","grey" );
    # now plot all tracks
    for iplot,groupname in enumerate(grouplist):
      plt = fig.add_subplot(nrow,2 if dual else 1,plotnums[iplot]);
      keys = list(self.groups[groupname]);
      nx = max([len(self.data.get(key,[])) for key in keys]);
      # first and last key
      key0 = keys[0];
      key1 = keys[-1];
      mindata,maxdata = 1e+99,-1e+99;
      # make plots for all ifrs
      for ikey,key in enumerate(keys):
        dd = self.data.get(key);
        if dd is None or dd.mask.all():
          continue;
        else:
          try:
            if len(dd) == 1:
              dd = numpy.array([dd[0],dd[0]]);
            plt.plot(dd,',',color=colors[ikey%len(colors)],label=self.label[key]);
          except:
            traceback.print_exc();
            print "Error plotting data for",key;
            continue;
      # add text labels
      # legend is not quite functional in matplotlib 0.98, so use text() instead
      y0,y1 = plt.get_ylim();
      if _version_major > 0 or _version_minor > 98:
        legend = plt.legend(ncol=len(keys),numpoints=1,prop=dict(size=5),handletextpad=0,columnspacing=0);
  #      legend.set_frame_on(False);
        for i,txt in enumerate(legend.get_texts()):
          txt.set_color(colors[i%len(colors)]);
        legend.get_frame().set_edgecolor('None');
        legend.get_frame().set_facecolor('white');
        legend.get_frame().set_alpha(.6);
      else:
        x0 = 0;
        for ikey,key in enumerate(keys):
          txt = plt.text(x0,y1,self.label[key],color=colors[ikey%len(colors)],
                          size=5,horizontalalignment='left',verticalalignment='top',
                          bbox=dict(facecolor='white',edgecolor='none',alpha=0.6));
          x0 += nx/20;
      plt.set_ylabel(groupname,size=8);
      # set x grid
      if _version_major > 0 or _version_minor > 98:
        plt.minorticks_on();
      for lab in plt.get_xmajorticklabels():
        lab.set_fontsize(5);
      for lab in plt.get_xminorticklabels():
        lab.set_fontsize(0);
      for lab in plt.get_ymajorticklabels():
        lab.set_fontsize(5);
      for lab in plt.get_yminorticklabels():
        lab.set_fontsize(0);
      if xgrid and xgrid != [0]:
        if isinstance(xgrid,(list,tuple)):
          if len(xgrid)>0:
            plt.set_xticks(range(0,nx,xgrid[0]));
          if len(xgrid)>1:
            plt.set_xticks(range(0,nx,xgrid[1]),True);
        for x in plt.get_xticks():
          plt.axvline(x,color='0.7',ls='-',lw=1,zorder=-10);
        for x in plt.get_xticks(True):
          plt.axvline(x,color='0.9',ls='-',lw=1,zorder=-10);
        plt.set_xlim(0,nx);
      if ygrid:
        for y in plt.get_yticks():
          plt.axhline(y,color='0.7',ls='-',lw=1,zorder=-10);
        for y in plt.get_yticks(True):
          plt.axhline(y,color='0.9',ls='-',lw=1,zorder=-10);
        plt.set_ylim(y0,y1);
        
    fig.subplots_adjust(left=mleft,right=1-mright,top=1-mtop,bottom=mbottom,wspace=0.12);
    # plot title if asked to
    if suptitle:
      fig.suptitle(suptitle,y=ytitle,size=8);
    if save:
      fig.savefig(save,papertype=papertype,dpi=dpi,
                  orientation='portrait' if not landscape else 'landscape');
      print "===> Wrote",save;
    return fig;


class ComplexCirclePlot (PlotCollection):
  """ComplexCirclePlot plots a complex circle plot""";
  # plot colors: xx xy yx yy
  colors = ("blue","red","purple","green");

  # function to make single-page plot
  def make_figure (self,keylist=None,suptitle=None,save=None,xgrid=False,ygrid=False,
                   dual=True,offset_std=10,
                   figsize=(210,290),dpi=100,papertype='a4',landscape=False):
    import matplotlib.pyplot as pyplot
    figsize_in = (figsize[0]/25.4,figsize[1]/25.4);
    fig = pyplot.figure(figsize=figsize_in,dpi=100);
    if keylist:
      datalist = [ (key,self.data.get(key)) for key in keylist ];
    else:
      datalist = data.iteritems();
    # margin sizes. We want to keep them fixed in absolute terms,
    # regardless of plot size. The relative numbers here are good for a 210x210 plot, so
    # we rescale them accordingly.
    mleft   = 0.06 * 210./figsize[0];
    mbottom = 0.06 * 210./figsize[1];
    mright  = 0.01 * 210./figsize[0];
    mtop    = 0.04 * 210./figsize[1];
    ytitle  = 1 - 0.01 * 210./figsize[1];
    width   = ( 1 - mleft - mright );
    height  = ( 1 - mtop - mbottom );
    plt = fig.add_axes([mleft,mbottom,width,height]);
    # find length of common prefix of labels
    labels = [lab for lab in self.label.itervalues() if lab];
    if labels:
      prefix = 1;
      while len(labels[0])>prefix and all([lab.startswith(labels[0][:prefix]) for lab in labels]):
        prefix += 1;
    # plot data over text labels
    ncorr = 0;
    rad = 0;
    for key,data in datalist:
      label = self.label.get(key);
      ncorr = max(ncorr,data.shape[-1]);
      for icorr in range(data.shape[-1]):
        color = self.colors[icorr];
        d = data[...,icorr];
        rad = max(rad,abs(d).max());
        if d.size > 1 or not label:
          plt.plot(d.real,d.imag,".-",mec=color,mfc=color,zorder=0);
        if label:
          d0 = d[-1] if d.ndim else d;
          plt.text(d0.real,d0.imag,label[prefix-1:],horizontalalignment='center',
                  verticalalignment='center',zorder=1,fontsize=8,color=color);
    # plot arrow to means
    for icorr in range(ncorr):
      color = self.colors[icorr];
      datameans = [d[...,icorr].mean() for d in self.data.itervalues() if d.shape[-1]>icorr ];
      meanval = sum(datameans)/len(datameans) if datameans else 0;
      plt.arrow(0,0,meanval.real,meanval.imag,color=color);
      # plot circle
      phi = numpy.arange(0,1.001,0.001)*2*math.pi;
      plt.plot(numpy.cos(phi)*abs(meanval),numpy.sin(phi)*abs(meanval),"-",color=color);
    for lab in plt.get_xticklabels():
      lab.set_fontsize(10);
    for lab in plt.get_yticklabels():
      lab.set_fontsize(10);
      lab.set_rotation('vertical');
    if xgrid or ygrid:
      plt.grid(color="0.4",linestyle='-');
    # plot title if asked to
    plt.set_xlim(-rad,rad);
    plt.set_ylim(-rad,rad);
    if suptitle:
      fig.suptitle(suptitle,y=ytitle,size=10);
    if save:
      fig.savefig(save,papertype=papertype,dpi=dpi,
                  orientation='portrait' if not landscape else 'landscape');
      print "===> Wrote",save;
    return fig;




class ScatterPlot (object):
  """ScatterPlot plots a complex scatterplot""";
  def __init__ (self,options):
    self.data = {};
    self.label = {};
    self.track_counts = {};
    # inter-plot offset -- may be set from outside
    self.offset = 0;

  def add_point (self,key,x,y,label=None):
    if isinstance(x,float):
      x = [x];
    if isinstance(y,float):
      y = [y];
    self.data[key] = x,y;
    self.label[key] = label if label is not None else str(key);

  def get_point (self,key):
    return self.data.get(key,None);

  # function to make single-page plot
  def make_figure (self,suptitle=None,save=None,xlabel=None,ylabel=None,
                   offset_std=10,xgrid=False,ygrid=False,
                   figsize=(210,210),dpi=100,papertype='a4',landscape=False):
    import matplotlib.pyplot as pyplot
    # create Figure object of given size and resolution
    figsize_in = (figsize[0]/25.4,figsize[1]/25.4);
    fig = pyplot.figure(figsize=figsize_in,dpi=100);
    # margin sizes. We want to keep them fixed in absolute terms,
    # regardless of plot size. The relative numbers here are good for a 210x210 plot, so
    # we rescale them accordingly.
    mleft   = 0.06 * 210./figsize[0];
    mbottom = 0.06 * 210./figsize[1];
    mright  = 0.01 * 210./figsize[0];
    mtop    = 0.04 * 210./figsize[1];
    ytitle  = 1 - 0.01 * 210./figsize[1];
    width   = ( 1 - mleft - mright );
    height  = ( 1 - mtop - mbottom );
    plt = fig.add_axes([mleft,mbottom,width,height]);
    # remove common prefix from labels
    if self.label:
      lab0 = self.label.values()[0];
      prefix = 1;
      while len(lab0) >= prefix and all([lab.startswith(lab0[:prefix]) for lab in self.label.itervalues()]):
        prefix += 1;
      labels = dict([(key,label[prefix-1:]) for key,label in self.label.iteritems()]);
    else:
      labels = None;
    # plot data over text labels
    for key,(x,y) in self.data.iteritems():
      # plot data
      try:
        plt.plot(x,y,"-x",zorder=0);
      except:
        traceback.print_exc();
        print "Error plotting data for",key;
        continue;
      # plot text labels
      if labels.get(key):
        plt.text(x[0],y[0]," "+labels[key],horizontalalignment='left',verticalalignment='top',zorder=1,fontsize=8);
#    plt.set_xbound(min([min(x) for x,y in self.data.itervalues()])[0],
#                   max([max(x) for x,y in self.data.itervalues()])[0]);
#    plt.set_ybound(min([min(y) for x,y in self.data.itervalues()])[0],
#                   max([max(y) for x,y in self.data.itervalues()])[0]);
    for lab in plt.get_xticklabels():
      lab.set_fontsize(10);
    for lab in plt.get_yticklabels():
      lab.set_fontsize(10);
      lab.set_rotation('vertical');
    # plot title if asked to
    if xlabel:
      plt.set_xlabel(xlabel,size=10);
    if ylabel:
      plt.set_ylabel(ylabel,size=10);
    if suptitle:
      fig.suptitle(suptitle,y=ytitle,size=10);
    if save:
      fig.savefig(save,papertype=papertype,dpi=dpi,
                  orientation='portrait' if not landscape else 'landscape');
      print "===> Wrote",save;
    return fig;


PLOT_SINGLE = 'single';
PLOT_MULTI  = 'multi';
PLOT_ERRORBARS = 'errorbars';
PLOT_BARPLOT = 'barplot';

class AbstractBasePlot (object):
  """Abstract base class for the various plotting objects""";

  def __init__ (self,options,figsize=(290,210)):
    init_pyplot(options.output_type);
    self.options = options;
    if options.width and options.height:
      self._default_figsize = self._user_figsize = (options.width,options.height);
    else:
      self._user_figsize = None;
      self._default_figsize = figsize;

  @classmethod
  def init_options (self,plotgroup,outputgroup):
    """Initializes options relevant to this plotting class""";
    self._plotgroup = plotgroup;
    self._outputgroup = outputgroup;

  @classmethod
  def add_plot_option (self,opt,*args,**kw):
    """Helper method: adds a plot option, if it's not already defined""";
    if not self._plotgroup.has_option(opt):
      self._plotgroup.add_option(opt,*args,**kw);

  @classmethod
  def add_output_option (self,opt,*args,**kw):
    """Helper method: adds an output option, if it's not already defined""";
    if not self._outputgroup.has_option(opt):
      self._outputgroup.add_option(opt,*args,**kw);

class MultigridPlot (AbstractBasePlot):
  """MultigridPlot makes NxM plots on a single page""";

  def __init__ (self,options,figsize=(290,210)):
    AbstractBasePlot.__init__(self,options,figsize);
    self._borders = [ 0.05,0.99,0.05,0.95 ];
    if options.borders and isinstance(options.borders,str):
      try:
        self._borders = [ float(x) for x in options.borders.split(",") ];
      except:
        pass;
    if options.y_lock:
      try:
        y0,y1 = map(float,options.y_lock.split(","));
      except:
        print "Error parsing --y-lock option, --y-lock MIN,MAX expected."
        sys.exit(1);
      self.ylock = y0,y1;
    else:
      self.ylock = None;

  @classmethod
  def init_options (self,plotgroup,outputgroup):
    AbstractBasePlot.init_options(plotgroup,outputgroup);
    self.add_plot_option("--title-fontsize",metavar="POINTS",type="int",default=12,
                    help="Set plot title font size, 0 for no title. Default is %default.");
    self.add_plot_option("--subtitle-fontsize",metavar="POINTS",type="int",default=10,
                    help="Set plot subtitle font size, 0 for none. Default is %default.");
    self.add_plot_option("--label-fontsize",metavar="POINTS",type="int",default=8,
                    help="Set plot label font size, 0 for no labels. Default is %default.");
    self.add_plot_option("--text-fontsize",metavar="POINTS",type="int",default=6,
                    help="Set plot text font size, 0 for no labels. Default is %default.");
    self.add_plot_option("--axis-fontsize",metavar="POINTS",type="int",default=5,
                    help="Set axis label font size, 0 for no axis labels. Default is %default.");
    self.add_plot_option("--text-spacing",metavar="VALUE",type="float",default=0.15,
                    help="Set spacing between lines of text in text comments. Default is %default.");
    self.add_plot_option("--y-ticks",metavar="N",type="int",default=4,
                    help="Maximum number of major ticks along the Y axis. Default is %default.");
    self.add_plot_option("--y-minor-ticks",metavar="INTERVAL",type="float",default=0,
                    help="Add minor ticks along the Y axis at the given intervals.");
    self.add_plot_option("--y-lock",metavar="MIN,MAX",type="str",default="",
                    help="Locks the Y scale of the plot to the given min/max values.");
    self.add_plot_option("--borders",metavar="L,R,B,T",type="str",
                    help="Set left/right/bottom/top plot borders, in normalized page coordinates. Default is %default.");
    self.add_plot_option("--subplot-wspace",metavar="W",type="float",default=0,
                    help="Set spacing between subplots, as fraction of subplot width. Default is automatic.");
    self.add_plot_option("-S","--subtitle",type="str",default="",
                      help="Subtitle for plot, added (in parentheses) after plot title");

    self.add_output_option("-o","--output-type",metavar="TYPE",type="string",default="png",
                      help="File format, see matplotlib documentation "
                      "for supported formats. At least 'png', 'pdf', 'ps', 'eps' and 'svg' are supported, or use 'x11' to display "
                      "plots interactively. Default is '%default.'");
    self.add_output_option("-r","--refresh",action="store_true",
                      help="Refresh plots even if they already exist (default is to keep existing plots.)");
    self.add_output_option("--output-prefix",metavar="PREFIX",type="string",default="",
                      help="Prefix output filenames with PREFIX_");
    self.add_output_option("--papertype",dest="papertype",type="string",default="a4",
                      help="set paper type (for .ps output only.) Default is '%default', but can also use e.g. 'letter', 'a3', etc.");
    self.add_output_option("-W","--width",type="int",default=0,
                      help="set explicit plot width, in mm. (Useful for .eps output)");
    self.add_output_option("-H","--height",type="int",default=0,
                      help="set explicit plot height, in mm. (Useful for .eps output)");
    self.add_output_option("--dpi",type="int",default=300,
                      help="figure resolution. Default is %default.");
    self.add_output_option("--scale",type="float",default=1,
                      help="scale plot sizes up by the given factor.");
    self.add_output_option("--portrait",action="store_true",
                    help="Force portrait orientation. Default is to select orientation based on plot size");
    self.add_output_option("--landscape",action="store_true",
                    help="Force landscape orientation.");

  @staticmethod
  def get_plot_data (data,xaxis=None):
    """Helper function, interprets the plot data returned by a datafunc.
    'data' is input data as returned by the datafunc argument to make_figure() below.
    'xaxis' is default X axis, Note to use ordinal numbering.
    Return value is tuple of X,Y,Yerr. Yerr is None if no error bars are provided.
    """;
    # a 2- or 3-tuple is interpreted as x,y[,yerr]
    if isinstance(data,tuple):
      if len(data) == 2:
        x,y = data;
        yerr = None;
      elif len(data) == 3:
        x,y,yerr = data;
      else:
        raise TypeError,"incorrect datum returned: 2- or 3-tuple expected, got %d-tuple"""%len(data);
    # else interpret data as Y
    else:
      y = data;
      x = yerr = None;
    # set X axis
    if x is None:
      x = range(len(y)) if xaxis is None else xaxis;
    # check lengths
    if len(x) != len(y):
      raise TypeError,"misshaped datum returned: %d X elements, %d Y elements"""%(len(x),len(y));
    if yerr is not None and len(yerr) != len(y):
      raise TypeError,"misshaped datum returned: %d Y elements, %d Yerr elements"""%(len(y),len(yerr));
    return x,y,yerr;

  def make_figure (self,rows,cols,  # (irow,row) and (icol,col) list
                  datafunc,   # datafunc(irow,icol) returns plot data for plot i,j
                  mode=PLOT_SINGLE,xaxis=None,
                  hline=None, # plot horizontal line at Y position, None for none
                  ylock="row", # lock Y scale. "row" locks across rows, "col" locks across columns, True locks across whole plot, (ymin,ymax) sets an explicit scale
                  mean_format="%.2f",
                  suptitle=None,      # title of plot
                  save=None,          # filename to save to
                  format=None,        # format: use options.output_type by default
                  figsize=None        # figure width,height in mm
                  ):
    from matplotlib import ticker
    if save and (format or self.options.output_type.upper()) != 'X11':
      save = "%s.%s"%(save,format or self.options.output_type);
      if self.options.output_prefix:
        save = "%s_%s"%(self.options.output_prefix,save);
      # exit if figure already exists, and we're not refreshing
      if os.path.exists(save) and not self.options.refresh: # and os.path.getmtime(save) >= os.path.getmtime(__file__):
        print save,"exists, not redoing";
        return save;
    else:
      save = None;

    figsize = numpy.array(self._user_figsize or figsize or self._default_figsize)/25.4;
    figsize *= self.options.scale;
    fig = pyplot.figure(figsize=figsize,dpi=10);
    iplot = 0;
    rows = list(rows);
    cols = list(cols);
    if ylock and not self.ylock:
      dummy = numpy.array([0.]);
      # form up ymin, ymax: NROWxNCOL arrays of min/max values per each plot
      if mode is PLOT_SINGLE or mode is PLOT_BARPLOT:
        datafilter = lambda x:x if isinstance(x,numpy.ndarray) else dummy;
        ymin = numpy.array([[datafilter(datafunc(row[0],col[0])).min() for col in cols ] for row in rows]);
        ymax = numpy.array([[datafilter(datafunc(row[0],col[0])).max() for col in cols ] for row in rows]);
      elif mode is PLOT_MULTI:
        ymin = numpy.array([[ min([self.get_plot_data(dd,xaxis)[1].min() for dd in datafunc(row[0],col[0])]) for col in cols ] for row in rows ]);
        ymax = numpy.array([[ max([self.get_plot_data(dd,xaxis)[1].max() for dd in datafunc(row[0],col[0])]) for col in cols ] for row in rows ]);
      elif mode is PLOT_ERRORBARS:
        dummy = numpy.array([0.]);
        datafilter1 = lambda x:x[0]-x[1] if isinstance(x[0],numpy.ndarray) else dummy;
        datafilter2 = lambda x:x[0]+x[1] if isinstance(x[0],numpy.ndarray) else dummy;
        ymin = numpy.array([[ datafilter1(datafunc(row[0],col[0])).min() for col in cols ] for row in rows]);
        ymax = numpy.array([[ datafilter2(datafunc(row[0],col[0])).max() for col in cols ] for row in rows]);
      # now collapse them according to ylock mode
      if isinstance(ylock,(list,tuple)) and len(ylock) == 2:
        ymin[...],ymax[...] = ylock;
      elif ylock == "row":
        ymin[...] = ymin.min(1)[:,numpy.newaxis];
        ymax[...] = ymax.max(1)[:,numpy.newaxis];
      elif ylock == "col":
        ymin[...] = ymin.min(0)[numpy.newaxis,:];
        ymax[...] = ymax.max(0)[numpy.newaxis,:];
      else:
        ymin[...] = ymin.min();
        ymax[...] = ymax.max();
    # make legend across the top
    for icol,col in cols:
      iplot += 1;
      plt = fig.add_subplot(len(rows)+1,len(cols)+1,iplot);
      plt.axis("off");
      plt.text(.5,0,col,fontsize=self.options.subtitle_fontsize*self.options.scale,
        horizontalalignment='center',verticalalignment='bottom');
    iplot += 1;
    # helper function to put lines of text instead of real plot
    def plot_text (plt,lines):
      y0 = 0.95;
      for i,text in enumerate(lines):
        plt.text(0,y0-self.options.text_spacing*i,text,
           fontsize=self.options.text_fontsize*self.options.scale,horizontalalignment='left',verticalalignment='top');
      plt.axis("off");
      plt.set_ylim(0,1);
    # now make plots
    for irow,row in rows:
      for icol,col in cols:
        iplot += 1;
        plt = fig.add_subplot(len(rows)+1,len(cols)+1,iplot);
        # plt.axis("off");
        plt.yaxis.set_major_locator(ticker.MaxNLocator(self.options.y_ticks));
        if self.options.y_minor_ticks:
          plt.yaxis.set_minor_locator(ticker.MultipleLocator(self.options.y_minor_ticks));
        if ylock and ylock != "col" and icol:
          plt.set_yticklabels([]);
        plt.set_xticks([]);
        data = datafunc(irow,icol);
        realplot = True;
        # if data is a sequence of strings, plot them
        if isinstance(data,(list,tuple)) and isinstance(data[0],str):
          plot_text(plt,data);
          realplot = False;
        elif mode is PLOT_SINGLE:
          plt.plot(range(len(data)) if xaxis is None else xaxis,data);
          if xaxis is None:
            plt.set_xlim(-1,len(data));
        elif mode is PLOT_BARPLOT:
          plt.bar(range(len(data)) if xaxis is None else xaxis,data,linewidth=0);
          if xaxis is None:
            plt.set_xlim(-1,len(data));
        elif mode is PLOT_MULTI:
          for dd in data:
            x,y,yerr = self.get_plot_data(dd,xaxis);
            if yerr is None:
              plt.plot(x,y);
            else:
              plt.errorbar(x,y,yerr,fmt=None,capsize=1);
            if x == range(len(y)):
              plt.set_xlim(-1,len(y));
        elif mode is PLOT_ERRORBARS:
          y,yerr = data;
          if len(y) == 1:
            plot_text(plt,(mean_format%y[0],"+/-"+(mean_format%yerr[0])));
            realplot = False;
          else:
            plt.errorbar(range(len(y)) if xaxis is None else xaxis,y,yerr,fmt=None,capsize=1);
            if xaxis is None:
              plt.set_xlim(-1,len(y));
        if realplot:
          if self.ylock:
            plt.set_ylim(*self.ylock);
            if icol:
              plt.set_yticklabels([]);
          elif ylock:
            plt.set_ylim(ymin[irow,icol],ymax[irow,icol]);
            if ylock is not "col" and icol:
              plt.set_yticklabels([]);
          if hline is not None:
            plt.axhline(y=hline,color='black');
        for lab in plt.get_yticklabels():
          lab.set_fontsize(self.options.axis_fontsize*self.options.scale);
      # add row label
      iplot += 1;
      plt = fig.add_subplot(len(rows)+1,len(cols)+1,iplot);
      plt.text(0,.5,row,fontsize=self.options.label_fontsize*self.options.scale,horizontalalignment='left',verticalalignment='center');
      plt.axis("off");
    # adjust layout
    fig.subplots_adjust(left=self._borders[0],right=self._borders[1],
      top=self._borders[3],bottom=self._borders[2],
      wspace=self.options.subplot_wspace or None);
    # plot title if asked to
    if suptitle and self.options.title_fontsize:
      if self.options.subtitle:
        suptitle = "%s (%s)"%(suptitle,self.options.subtitle);
      fig.suptitle(suptitle,fontsize=self.options.title_fontsize*self.options.scale);
    if save:
      if self.options.portrait:
        orientation = 'portrait';
      elif self.options.landscape:
        orientation = 'landscape';
      else:
        orientation='portrait' if figsize[0]<figsize[1] else 'landscape';
      fig.savefig(save,papertype=self.options.papertype,dpi=self.options.dpi,
                  orientation=orientation);
      print "Wrote",save,"in",orientation,"orientation";
      fig = None;
      pyplot.close("all");
    return save;


class SkyPlot (AbstractBasePlot):
  """SkyPlot plots things in an l,m plot""";
  def __init__ (self,options,figsize=(290,210)):
    AbstractBasePlot.__init__(self,options,figsize);
    self._borders = [ .05,.9,.05,.9 ];
    if options.circle_borders and isinstance(options.circle_borders,str):
      try:
        self._borders = [ float(x) for x in options.circle_borders.split(",") ];
      except:
        pass;

  @classmethod
  def init_options (self,plotgroup,outputgroup):
    AbstractBasePlot.init_options(plotgroup,outputgroup);
    self.add_plot_option("--radius",type="int",default=0,
                      help="radius of circle plots, in arcmin. Default is auto-sizing.");
    self.add_plot_option("-G","--grid-circle",metavar="ARCMIN",type="int",action="append",
                      help="sets radius of grid circle(s) in circle plots. May be given multiple times. Default is 30 and 60.");
    self.add_plot_option("--circle-title-fontsize",metavar="POINTS",type="int",default=12,
                    help="Set plot title font size in circle plots, 0 for no title. Default is %default.");
    self.add_plot_option("--circle-label-fontsize",metavar="POINTS",type="int",default=8,
                    help="Set plot label font size in circle plots, 0 for no labels. Default is %default.");
    self.add_plot_option("--circle-axis-fontsize",metavar="POINTS",type="int",default=5,
                    help="Set axis label font size in circle_plots, 0 for no axis labels. Default is %default.");
    self.add_plot_option("--names-only",action="store_true",
                    help="Use only source names for labels in circle plots. Default is name, mean and std values.");
    self.add_plot_option("--circle-label-offset",metavar="FRAC",type="float",default=0.02,
                    help="Vertical offset of circle labels. Use 0 to center labels on source position. Default is %default.");
    self.add_plot_option("--circle-borders",metavar="L,R,B,T",type="str",
                    help="Set left/right/bottom/top circle-plot borders, in normalized page coordinates. Default is %default.");
    self.add_plot_option("-S","--subtitle",type="str",default="",
                      help="Subtitle for plot, added (in parentheses) after plot title");

    self.add_output_option("-o","--output-type",metavar="TYPE",type="string",default="png",
                    help="File format, see matplotlib documentation "
                    "for supported formats. At least 'png', 'pdf', 'ps', 'eps' and 'svg' are supported, or use 'x11' to display "
                    "plots interactively. Default is '%default.'");
    self.add_output_option("-r","--refresh",action="store_true",
                      help="Refresh plots even if they already exist (default is to keep existing plots.)");
    self.add_output_option("--output-prefix",metavar="PREFIX",type="string",default="",
                      help="Prefix output filenames with PREFIX_");
    self.add_output_option("--papertype",dest="papertype",type="string",default="a4",
                      help="set paper type (for .ps output only.) Default is '%default', but can also use e.g. 'letter', 'a3', etc.");
    self.add_output_option("-W","--width",type="int",
                      help="set explicit plot width, in mm. (Useful for .eps output)");
    self.add_output_option("-H","--height",type="int",
                      help="set explicit plot height, in mm. (Useful for .eps output)");
    self.add_output_option("--dpi",type="int",default=300,
                      help="figure resolution. Default is %default.");
    self.add_output_option("--portrait",action="store_true",
                    help="Force portrait orientation. Default is to select orientation based on plot size");
    self.add_output_option("--landscape",action="store_true",
                    help="Force landscape orientation.");

  """Plotting function""";
  def make_figure (self,ll,mm,
    markers,  # marker is a list of strings or dicts or tuples.
              # For each marker, if str: axes.plot(x,y,str) is called
              # if dict: axes.plot(x,y,**dict)
              # else tuple is func,arg,kwargs
              # and axes.func(*args,**kwargs) is called
    labels=None,
    radius=None,        # plot radius. If None, then it's set automatically
    zero_lines=True,    # plot l=0 and m=0 lines
    suptitle=None,      # title of plot
    save=None,          # filename to save to
    format=None,        # format: use self.options.output_type by default
    figsize=(210,210),       # figure width,height in mm
    ):
    if save and (format or self.options.output_type.upper()) != 'X11':
      save = "%s.%s"%(save,format or self.options.output_type);
      if self.options.output_prefix:
        save = "%s_%s"%(self.options.output_prefix,save);
      # exit if figure already exists, and we're not refreshing
      if os.path.exists(save) and not self.options.refresh: # and os.path.getmtime(save) >= os.path.getmtime(__file__):
        print save,"exists, not redoing";
        return save;
    else:
      save = None;

    figsize = numpy.array(figsize or self._default_figsize)/25.4;
    fig = pyplot.figure(figsize=figsize,dpi=600);
    plt = fig.add_axes([self._borders[0],self._borders[2],self._borders[1]-self._borders[0],
                        self._borders[3]-self._borders[2]]);
    if zero_lines:
      plt.axhline(y=0,color='black',linestyle=':');
      plt.axvline(x=0,color='black',linestyle=':');

    # if ll is None, use length of markers for everything
    if ll is None:
      ll = mm = numpy.zeros(len(markers),float);

    if not isinstance(markers,(list,tuple)):
      markers = [markers]*len(ll);

    if not labels:
      labels = [None]*len(ll);

    # set limits
    if self.options.radius:
      lim = self.options.radius;
    elif radius:
      lim = radius;
    else:
      lim = max(max(abs(ll)),max(abs(mm)))*1.05/ARCMIN;

    # plot markers
    for l,m,mark,label in zip(ll/ARCMIN,mm/ARCMIN,markers,labels):
      if isinstance(mark,str):
        plt.plot(l,m,mark);
      elif isinstance(mark,dict):
        plt.plot(l,m,**mark);
      elif isinstance(mark,tuple):
        func,args,kwargs = mark;
        getattr(plt,func)(*args,**kwargs);
      if label and self.options.circle_label_fontsize:
        plt.text(l,m-lim*self.options.circle_label_offset,
          label,fontsize=self.options.circle_label_fontsize,
          horizontalalignment='center',
          verticalalignment='top' if self.options.circle_label_offset else 'center');

    x = numpy.cos(numpy.arange(0,361)*math.pi/180);
    y = numpy.sin(numpy.arange(0,361)*math.pi/180);

    gc = self.options.grid_circle or (30,60);
    for R in gc:
      plt.plot(x*R,y*R,linestyle=':',color='black');

    plt.set_xlim(lim,-lim);
    plt.set_ylim(-lim,lim);
    for lab in list(plt.get_xticklabels())+list(plt.get_yticklabels()):
      lab.set_fontsize(self.options.circle_axis_fontsize);

    if suptitle and self.options.title_fontsize:
      if self.options.subtitle:
        suptitle = "%s (%s)"%(suptitle,self.options.subtitle);
      fig.suptitle(suptitle,fontsize=self.options.circle_title_fontsize);
    if save:
      if self.options.portrait:
        orientation = 'portrait';
      elif self.options.landscape:
        orientation = 'landscape';
      else:
        orientation='portrait' if figsize[0]<figsize[1] else 'landscape';
      fig.savefig(save,papertype=self.options.papertype,dpi=self.options.dpi,
                  orientation=orientation);
      print "Wrote",save,"in",orientation,"orientation";
      fig = None;
      pyplot.close("all");
    return save;




#!/usr/bin/python
# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import os.path
import re
import time
import math
import Owlcat

ROWCHUNK = 50000;

time_start = time.time();

import pyrap.tables
from pyrap.tables import table,tablecopy,tableexists,tabledelete,tablecreatearraycoldesc

def timestamp (format="%H:%M:%S:"):
  return time.strftime(format,time.gmtime(time.time()-time_start));

def progress (message,newline=True):
  sys.stdout.write("%s %-60s%s"%(timestamp(),message,'\n' if newline else '\r'));
  sys.stdout.flush();

def call_system (command,args):
  rc = os.system("%s %s"%(command,args));
  if rc:
    stat = rc>>8;
    sig = rc&0xFF;
    if sig:
      print "%s process killed by signal %d"%(command,sig);
    else:
      print "%s call failed with exit status",(command,stat);
      if stat == 127:
        print "Please check your MeqTrees installation for the %s utility!"%command;
  return rc;
  


if __name__ == "__main__":

  # setup some standard command-line option parsing
  #
  from optparse import OptionParser
  parser = OptionParser(usage="""%prog: [options] MS""",
      description=\
      "Sets up tiled storage managers for the DATA, FLAG and BITFLAG columns of an MS, "+
      "and sets a fixed shape on the columns. This can speed up access to some MSs, e.g. "+
      "such as those produced by the uvfits2ms converter. "
  );
  parser.add_option("-f","--force",action="store_true",
                    help="proceed non-interactively");

  (options,msnames) = parser.parse_args();

  if len(msnames) != 1:
    parser.error("MS name not supplied.");
    
  if not Owlcat.find_exec("addtiledmscol"):
    print "Can't find the 'addtiledmscol' utility, which is required by this script."
    print "You may need to install or build the MeqTrees package.";
    sys.exit(1);

  msname = msnames[0];
  ms = table(msname,readonly=False);

  if not options.force:
    print """
  This script will reinitialize the DATA, FLAG and BITFLAG columns
  of the MS using a tiled storage manager and a fixed shape. It will attempt
  to preserve the data in these columns. The MODEL_DATA, CORRECTED_DATA and
  IMAGING_WEIGHT columns will be reset and their contents discarded. If this
  operation is interrupted, the MS may be left in a corrupted state, so
  please make sure you have a backup of the MS.
  """;
    if raw_input("Proceed (y/n)? ").strip().upper()[0] != "Y":
      print "Aborted by user.";
      sys.exit(1);

  SAVECOLS = [ "DATA","FLAG","BITFLAG" ];
  KILLCOLS = SAVECOLS + [ "MODEL_DATA","CORRECTED_DATA","IMAGING_WEIGHT" ];

  # read in DATA and FLAG cols
  savedcols = {};
  for colname in "DATA","FLAG","BITFLAG":
    if colname in ms.colnames():
      progress("Reading %s column"%colname);
      savedcols[colname] = ms.getcol(colname);
    elif colname is "DATA":
      print "Error: this MS does not appear to have a DATA column";
      sys.exit(1);

  datacol = savedcols['DATA'];
  progress("DATA column shape is %s"%(datacol.shape,));
  flagcol = savedcols.get('FLAG');
  if flagcol is not None and datacol.shape != flagcol.shape:
    print "Error: FLAG shape (%s) does not match DATA shape"%(flagcol.shape,);

  # kill columns
  cols = [ col for col in KILLCOLS if col in ms.colnames() ];
  if cols:
    progress("Removing columns %s"%", ".join(cols));
    ms.removecols(cols);

  # now reinsert DATA and FLAG columns
  ms.close();
  ncorr = datacol.shape[-1];
  nfreq = datacol.shape[-2];
  tilerow = 512;
  for colname,coltype in ("DATA","complex"),("FLAG","bool"),("BITFLAG","int"):
    if colname in savedcols:
      progress("Inserting tiled %s column"%colname);
      if not call_system("addtiledmscol","%s %s %s %d %d %d %d %d 4"%
                          (msname,colname,coltype,ncorr,nfreq,min(ncorr,4),min(nfreq,8),tilerow)):
        sys.exit(1);

  # reopen table
  ms = table(msname,readonly=False);
  for colname in SAVECOLS:
    col = savedcols.get(colname);
    if col is not None:
      progress("Rewriting %s column"%colname);
      ms.putcol(colname,col);

  progress("Closing MS");
  ms.close();
  progress("Adding imaging columns");
  pyrap.tables.addImagingColumns(msname);
  progress("Done");



# -*- coding: utf-8 -*-

#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import re
import os.path
import numpy
## ugly hack to get around UGLY FSCKING ARROGNAT (misspelling fully intentional) pyfits-2.3 bug
import Kittens.utils
pyfits = Kittens.utils.import_pyfits();
import scipy.ndimage.measurements
import math
from astLib.astWCS import WCS

SANITIZE_DEFAULT = 12345e-7689


def stack_planes(fitslist, outname='combined.fits', axis=0, ctype=None, keep_old=False, fits=False):
    """ Stack a list of fits files along a given axiis.
       
       fitslist: list of fits file to combine
       outname: output file name
       axis: axis along which to combine the files
       fits: If True will axis FITS ordering axes
       ctype: Axis label in the fits header (if given, axis will be ignored)
       keep_old: Keep component files after combining?
    """

    hdu = pyfits.open(fitslist[0])[0]
    hdr = hdu.header
    naxis = hdr['NAXIS']

    # find axis via CTYPE key
    if ctype is not None:
        for i in range(1,naxis+1):
            if hdr['CTYPE%d'%i].lower().startswith(ctype.lower()):
                axis = naxis - i # fits to numpy convention
    elif fits:
      axis = naxis - axis

    fits_ind = abs(axis-naxis)
    crval = hdr['CRVAL%d'%fits_ind]

    imslice = [slice(None)]*naxis
    _sorted = sorted([pyfits.open(fits) for fits in fitslist],
                    key=lambda a: a[0].header['CRVAL%d'%(naxis-axis)])

    # define structure of new FITS file
    nn = [ hd[0].header['NAXIS%d'%(naxis-axis)] for hd in _sorted]
    shape = list(hdu.data.shape)
    shape[axis] = sum(nn)
    data = numpy.zeros(shape,dtype=float)

    for i, hdu0 in enumerate(_sorted):
        h = hdu0[0].header
        d = hdu0[0].data
        imslice[axis] = range(sum(nn[:i]),sum(nn[:i+1]) )
        data[imslice] = d
        if crval > h['CRVAL%d'%fits_ind]:
            crval =  h['CRVAL%d'%fits_ind]

    # update header
    hdr['CRVAL%d'%fits_ind] = crval
    hdr['CRPIX%d'%fits_ind] = 1

    pyfits.writeto(outname, data, hdr, clobber=True)
    print("Successfully stacked images. Output image is %s"%outname)

    # remove old files
    if not keep_old:
        for fits in fitslist:
            os.system('rm -f %s'%fits)


def unstack_planes(fitsname, each_chunk, axis=None, ctype=None, prefix=None,fits=False):
    """ 
        Unstack FITS image along a given axis into multiple 
        images each having each_chunk planes along that axis 
    """

    prefix = prefix or fitsname[:-5] # take everthing but .FITS/.fits
    hdu = pyfits.open(fitsname)
    hdr = hdu[0].header
    data = hdu[0].data.copy()
    naxis = hdr["NAXIS"]

    if axis is None and ctype is None:
        raise SystemExit('Please specify either axis or ctype')
    # find axis via CTYPE key

    if ctype :
        for i in range(1,naxis+1):
            if hdr['CTYPE%d'%i].lower().startswith(ctype.lower()):
                axis = naxis - i # fits to numpy indexing
    elif fits:
      axis = naxis - axis

    crval = hdr['CRVAL%d'%(naxis-axis)]
    cdelt = hdr['CDELT%d'%(naxis-axis)]
    crpix = hdr['CRPIX%d'%(naxis-axis)]
    # shift crval to crpix=1
    crval = crval - (crpix-1)*cdelt

    nstacks = hdr['NAXIS%d'%(naxis-axis)]
    nchunks = nstacks//each_chunk
    print("The FITS file %s has %s stacks along this axis. Breaking it up to %d images"%(fitsname, nstacks, nchunks))

    outfiles = []
    for i in range(0, nchunks):

        _slice = [slice(None)]*naxis
        _slice[axis] = range(i*each_chunk, (i+1)*each_chunk if i+1!=nchunks else nstacks)
        hdu[0].data = data[_slice]
        hdu[0].header['CRVAL%d'%(naxis-axis)] = crval + i*cdelt*each_chunk
        hdu[0].header['CRPIX%d'%(naxis-axis)] = 1
        outfile = '%s-%04d.fits'%(prefix, i)
        outfiles.append(outfile)
        print("Making chunk %d : %s. File is %s"%(i, repr(_slice[axis]), outfile))
        hdu.writeto(outfile, clobber=True)
    hdu.close()
    return outfiles


def reorder(fitsname, order=[], outfile=None):
    """ 
    Re-order FITS image axes. 

    Example:
        If your FITS files has axes (NAXIS{1,2,3,4})-> RA, DEC, STOKES, FREQ
        You can re-order it so that it has RA, DEC, FREQ, STOKES
        by running: reorder(fitsname, order=[1,2,4,3], "my_reodered_image.fits")
    """

    # Get to know input FITS image
    hdu = pyfits.open(fitsname)
    hdr0 = hdu[0].header
    ndim = hdr0["NAXIS"]
    order0 = range(1,ndim+1)

    def fits2py(a):
        if isinstance(a, (list, tuple)):
            return list(ndim - numpy.array(a) )
        else:
            return ndim -a


    if order == order0:
        print("The FITS image already has the order you requested")
        return 
    elif len(order)!=ndim:
        raise ValueError("The dimensions of the FITS image do not match your request.")

    # Ok, Lock and Load
    data = hdu[0].data
    # First re-order data
    hdu[0].data = numpy.transpose(data, list((ndim-numpy.array(order))[range(ndim-1,-1,-1)]) )

    hdr = hdr0.copy()
    mendatory = "CTYPE CRVAL CDELT CRPIX".split()
    optional = "CUNIT CROTA".split()

    for key in mendatory+optional:
        for old, new in zip(order0,order):
            if old == new:
                continue
            try:
                val = hdr0["%s%d"%(key,old)]
                comment = hdr0.comments["%s%d"%(key,old)]
                idx = hdr.index( "%s%d"%(key,old) )
                # Seems like can't replace non empy val with an empty string. So delete first

                del hdr["%s%d"%(key,new)]
                hdr.insert(idx, ("%s%d"%(key,new), val, comment) )

            except KeyError: 
                if key in mendatory:
                    raise KeyError("ARBORTNG: FITS file doesn't have the '%s' key in the header"%key)
                else:
                    pass

    hdu[0].header = hdr
    outfile = outfile or "reodered_"+fitsname
    print("Successfully re-ordered axes in FITS image. Output image is at %s"%outfile)
    hdu.writeto(outfile, clobber=True)
    

def main():

  # setup some standard command-line option parsing
  #
  from optparse import OptionParser
  parser = OptionParser(usage="""%prog: [options] <image names...>""");
  parser.add_option("-o","--output",dest="output",type="string",
                    help="name of output FITS file");
  parser.add_option("-r","--replace",action="store_true",
                    help="replace (first) input file by output. Implies '--force'.");
  parser.add_option("-f","--force",dest="force",action="store_true",
                    help="overwrite output file even if it exists");
  parser.add_option("-S","--sanitize",type="float",metavar="VALUE",
                    help="sanitize FITS files by replacing NANs and INFs with VALUE");
  parser.add_option("-N","--nonneg",action="store_true",
                    help="replace negative values by 0");
  parser.add_option("-m","--mean",dest="mean",action="store_true",
                    help="take mean of input images");
  parser.add_option("-d","--diff",dest="diff",action="store_true",
                    help="take difference of 2 input images");
  parser.add_option("-t","--transfer",action="store_true",
                    help="transfer data from image 2 into image 1, preserving the FITS header of image 1");
  parser.add_option("-z","--zoom",dest="zoom",type="int",metavar="NPIX",                    
                    help="zoom into central region of NPIX x NPIX size");
  parser.add_option("-R","--rescale",dest="rescale",type="float",
                    help="rescale image values");
  parser.add_option("-E","--edit-header",metavar="KEY=VALUE",type="string",action="append",
                    help="replace header KEY with VALUE. Use KEY=VALUE for floats and KEY='VALUE' for strings.");
  parser.add_option("--stack",metavar="outfile:axis",
                    help="Stack a list of FITS images along a given axis. This axis may given as an integer"
                    "(as it appears in the NAXIS keyword), or as a string (as it appears in the CTYPE keyword)")
  parser.add_option("--reorder",
                    help="Required order. List of comma seperated indeces")
  parser.add_option("--add-axis",metavar="CTYPE:CRVAL:CRPIX:CDELT[:CUNIT:CROTA]",action="append",default=[],
                    help="Add axis to a FITS image. The AXIS will be described by CTYPE:CRVAL:CRPIX:CDELT[:CUNIT:CROTA]. "
                    "The keywords in brackets are optinal, while those not in brackets are mendatory. "
                    "This axis will be the last dimension. Maybe specified more than once.")
  parser.add_option("--unstack",metavar="prefix:axis:each_chunk",
                    help="Unstack a FITS image into smaller chunks each having [each_chunk] planes along a given axis. "
                    "This axis may given as an integer (as it appears in the NAXIS keyword), or as a string "
                    "(as it appears in the CTYPE keyword)")
  parser.add_option("-H","--header",action="store_true",help="print header(s) of input image(s)");
  parser.add_option("-s","--stats",action="store_true",help="print stats on images and exit. No output images will be written.");

  parser.set_defaults(output="",mean=False,zoom=0,rescale=1,edit_header=[]);

  (options,imagenames) = parser.parse_args();
  
  if not imagenames:
    parser.error("No images specified. Use '-h' for help.");

  # print "%d input image(s): %s"%(len(imagenames),", ".join(imagenames));
  images = [ pyfits.open(img) for img in imagenames ];
  updated = False;

  # Stack FITS images
  if options.stack:
    if len(imagenames)<1:
      parser.error("Need more than one image to stack")
    stack_args = options.stack.split(":")
    if len(stack_args) != 2:
      parser.error("Two --stack options are required. See ./fitstool.py -h")
   
    outfile,axis = stack_args

    _string = True
    try:
      axis = int(axis)
      _string = False
    except ValueError:
      _string = True

    stack_planes(imagenames,ctype=axis if _string else None,keep_old=True,
                 axis=None if _string else axis,outname=outfile,fits=True)
  
  # Unstack FITS image
  if options.unstack:
    image = imagenames[0]
    if len(imagenames)<1:
      parser.error("Need more than one image to stack")
    unstack_args = options.unstack.split(":")
    if len(unstack_args) != 3:
      parser.error("Two --unstack options are required. See ./fitstool.py -h")

    prefix,axis,each_chunk = unstack_args
    
    _string = True
    try:
      axis = int(axis)
      _string = False
    except ValueError:
      _string = True

    each_chunk = int(each_chunk)
    
    unstack_planes(image,each_chunk,ctype=axis if _string else None,
            axis=None if _string else axis,prefix=prefix,fits=True)
    sys.exit(0)
  
  if options.add_axis:
    for axis in options.add_axis:
      hdu = pyfits.open(imagenames[0])
      hdr = hdu[0].header
      ndim = hdr["NAXIS"]
      hdu[0].data = hdu[0].data[numpy.newaxis,...]
      _mendatory = "CTYPE CRVAL CDELT CRPIX".split()
      _optional = "CUNIT CROTA".split()
      L = len(_mendatory+_optional)
      values = axis.split(":")
      Lv = len(values)

      if len(_mendatory)>len(values) :
        parser.error("Something with the way specified --add-axis. See %prog -h for help")

      for i,value in enumerate(values ) :
        try:
          value = float(value)
        except ValueError:
          if isinstance(value, str):
            value = value.upper()

        hdu[0].header["%s%d"%( (_mendatory+_optional)[i],ndim+1)] = value

      hdu.writeto(imagenames[0],clobber=True)
      print("Successfully added axis %s to %s"%(values[0],imagenames[0]))
        

  if options.reorder:
    for image in imagenames:
      order = map(int, options.reorder.split(","))
      reorder(image, order=order, outfile=image)
  
  if options.header:
    for filename,img in zip(imagenames,images):
      if len(imagenames)>1:
        print "======== FITS header for",filename;
      for hdrline in img[0].header.ascard:
        print hdrline; 
  
  if options.replace or len(imagenames)<2:
    if options.output:
      parser.error("Cannot combine -r/--replace with -o/--output");
    outname = imagenames[0];
    options.force = True;
    autoname = False;
  else:
    outname = options.output;
    autoname = not outname;
    if autoname:
      outname = re.split('[_]',imagenames[0],1)[-1];

  for keyval in options.edit_header:
    key,val = keyval.split("=");
    q = ''
    if val[0] == "'" and val[-1] == "'":
      images[0][0].header[key] = val[1:-1:];
      q = '"'
    elif val[-1] == 'd' or key.startswith('NAXIS'):
      images[0][0].header[key] = int(val[:-1] if val[-1]=='d' else val);
    else:
      try:
        images[0][0].header[key] = float(val);
      except:
        images[0][0].header[key] = val;
        q = '"';
    print "Setting header %s=%s%s%s"%(key,q,val,q);
    updated = True;

  if options.sanitize is not None:
    print "Sanitizing: replacing INF/NAN with",options.sanitize;
    for img in images:
      d = img[0].data;
      d[numpy.isnan(d)+numpy.isinf(d)] = options.sanitize;
    # if using stats, do not generate output
    if not options.stats:
      updated = True;

  if options.nonneg:
    print "Replacing negative value by 0";
    for img,name in zip(images,imagenames)[:1]:
      d = img[0].data;
      wh = d<0;
      d[wh] = 0;
      print "Image %s: replaced %d points"%(name,wh.sum());
    updated = True;

  if options.transfer:
    if len(images) != 2:
      parser.error("The --transfer option requires exactly two input images.");
    if autoname:
      outname = "xfer_" + outname;
    print "Transferring %s into coordinate system of %s"%(imagenames[1],imagenames[0]);
    images[0][0].data = images[1][0].data;
    updated = True;
  elif options.diff:
    if len(images) != 2:
      parser.error("The --diff option requires exactly two input images.");
    if autoname:
      outname = "diff_" + outname;
    print "Computing difference";
    data = images[0][0].data;
    data -= images[1][0].data;
    updated = True;
  elif options.mean:
    if autoname:
      outname = "mean%d_"%len(images) + outname;
    print "Computing mean";
    data = images[0][0].data;
    for img in images[1:]:
      data += img[0].data;
    data /= len(images);
    images = [ images[0] ];
    updated = True;

  if options.zoom:
    z = options.zoom;
    if autoname:
      outname = "zoom%d_"%z + outname;
    if len(images) > 1:
      "Too many input images specified for this operation, at most 1 expected";
      sys.exit(2);
    data = images[0][0].data;
    nx = data.shape[-2];
    ny = data.shape[-1];
    zdata = data[:,:,(nx-z)/2:(nx+z)/2,(ny-z)/2:(ny+z)/2];
    #update header
    hdr = images[0][0].header
    wcs = WCS(hdr, mode="pyfits")
    cr1, cr2 = wcs.pix2wcs(ny/2, nx/2)
    hdr["CRVAL1"] = cr1
    hdr["CRVAL2"] = cr2 
    hdr["CRPIX1"] = zdata.shape[-1]/2
    hdr["CRPIX2"] = zdata.shape[-2]/2
    
    print "Making zoomed image of shape","x".join(map(str,zdata.shape));
    images = [ pyfits.PrimaryHDU(zdata, hdr) ];
    updated = True;

  if options.rescale != 1:
    if autoname and not updated:
      outname = "rescale_" + outname;
    if len(images) > 1:
      "Too many input images specified for this operation, at most 1 expected";
      sys.exit(2);
    print "Applying scaling factor of %f to image values"%options.rescale;
    images[0][0].data *= options.rescale;
    updated = True;

  if updated:
    imagenames[0] = outname;

  if options.stats:
    for ff,filename in zip(images,imagenames):
      data = ff[0].data;
      min,max,dum1,dum2 = scipy.ndimage.measurements.extrema(data);
      sum = data.sum();
      mean = sum/data.size;
      std = math.sqrt(((data-mean)**2).mean());
      print "%s: min %g, max %g, sum %g, np %d, mean %g, std %g"%(filename,min,max,sum,data.size,mean,std); 
    sys.exit(0);
      
  if updated:
    print "Writing output image",outname;
    if os.path.exists(outname) and not options.force:
      print "Output image exists, rerun with the -f switch to overwrite.";
      sys.exit(1);
    images[0].writeto(outname,clobber=True);
  elif not (options.header or options.stack or options.add_axis or options.reorder):
    print "No operations specified. Use --help for help."
